from os import path
from .. import DATA_DIR

DEFAULT_CONFIGURATION = {'address': '',
                         'plugin_http_port': 80,
                         'plugin_https_port': 443,
                         'vasa_provider_https_port': 8443,
                         'registered_ip_address': None,
                         'webserver_certificate': path.join(DATA_DIR, 'webserver.pem'),
                         'webserver_certificate_private_key': path.join(DATA_DIR, 'webserver.key'),
                         'vcenter_extension_certificate': path.join(DATA_DIR, 'vcenter_extension.pem'),
                         'vcenter_extension_certificate_private_key': path.join(DATA_DIR, 'vcenter_extension.key'),
                         'root_certificates': path.join(DATA_DIR, "root_certificates"),
                         'vcenter_extension_key': 'com.infinidat.powertools',
                         'in_production': True,
                         'active_session_count': 30,
                         'rotating_logfile_max_size': 1024*1024*20,
                         'rotating_logfile_backup_count': 10,
                         'vasa_provider_username': 'inyerzajlarwuokti',
                         'vasa_provider_password': 'iknockapsusheelco',
                         }

PRUDCTION_CONFIGURATION = DEFAULT_CONFIGURATION.copy()
DEVELOPMENT_CONFIGURATION = DEFAULT_CONFIGURATION.copy()

def _get_development_extension_key():
    from socket import gethostname
    hostname = gethostname().split('.')[0]
    production_key = DEFAULT_CONFIGURATION['vcenter_extension_key']
    kwargs = dict(production_key=production_key, hostname=hostname)
    return "{production_key}.develop.{hostname}".format(**kwargs)

DEVELOPMENT_CONFIGURATION.update({'plugin_http_port': 8080,
                                  'plugin_https_port': 8081,
                                  'vasa_provider_https_port': 8000,
                                  'vcenter_extension_key': _get_development_extension_key()})

CERTIFICATE_SUBJECT_NAME = '/C=IL/ST=Infinidat/O=Infinidat/CN={ip}'

from ..redis import RedisServer
from redis.exceptions import BusyLoadingError
from logging import getLogger
import cjson
logger = getLogger(__name__)

class Configuration(object):
    @classmethod
    def _get(cls, name):
        from redis import ConnectionError
        try:
            # If the server was stopped, and did not start yet, we should restart it
            # This happens during reload
            s = RedisServer.get_client().get(name)
            if isinstance(s, BusyLoadingError):  # retry
                s = RedisServer.get_client().get(name)
            return getattr(cjson, 'decode')(s or 'null')
        except ConnectionError:  # pragma: no cover
            RedisServer.start()
            return cls._get(name)

    @classmethod
    def _set(cls, name, value):
        RedisServer.get_client().set(name, getattr(cjson, 'encode')(value))

    @classmethod
    def _setnx(cls, name, value):
        RedisServer.get_client().setnx(name, getattr(cjson, 'encode')(value))

    @classmethod
    def populate_defaults(cls, production=True):
        cls.populate_default_configuration_to_redis(production)
        cls._set('in_production', production)
        if cls._get_source_address() != cls._get('registered_ip_address'):
            logger.info("Generating self-signed certificates")
            cls.generate_self_signed_certificates()
            cls._set('registered_ip_address', cls._get_source_address())
        logger.info("Configuration loaded")

    @classmethod
    def populate_default_configuration_to_redis(cls, production=True):
        logger.info("Populating default configuration in redis")
        conf = PRUDCTION_CONFIGURATION if production else DEVELOPMENT_CONFIGURATION
        for key, value in conf.items():
            cls._setnx(key, value)

    @classmethod
    def is_in_production(cls):
        return cls._get('in_production')

    @classmethod
    def get_certificate_subject_name(cls):
        return CERTIFICATE_SUBJECT_NAME.format(ip=cls._get_source_address())

    @classmethod
    def _openssl(cls, *args, **kwargs):
        from infi.execute import execute_assert_success
        cmd = ['openssl']
        cmd.extend(args)
        execute_assert_success(cmd)

    @classmethod
    def _generate_certificate(cls, certificate_filepath, private_key_filepath):
        # VASA Programming Guide 1.1, appendix A
        # http://longwhiteclouds.com/2012/02/07/the-trouble-with-ca-ssl-certificates-and-vcenter-5/
        private_key_with_passphrase = '.'.join([private_key_filepath, 'passphase'])
        request_filepath = '.'.join([certificate_filepath, 'csr'])
        extension_filepath = '.'.join([certificate_filepath, 'ext'])
        passphrase = 'pass:testpassword'
        with open(extension_filepath, 'w') as fd:
            fd.write('\n'.join([
                     "basicConstraints       = CA:FALSE",
                     "keyUsage               = nonRepudiation,digitalSignature,keyEncipherment,dataEncipherment",
                     "extendedKeyUsage       = serverAuth,clientAuth",
                     "subjectKeyIdentifier   = hash",
                     "authorityKeyIdentifier = keyid,issuer",
                     ""]))
        cls._openssl('genrsa', '-aes128', '-passout', passphrase,
                     '-out', private_key_with_passphrase, '1024')
        cls._openssl('rsa', '-passin', passphrase,
                     '-in', private_key_with_passphrase, '-out', private_key_filepath)
        cls._openssl('req', '-new', '-days', '36510', '-subj', cls.get_certificate_subject_name(),
                     '-key', private_key_filepath, '-out', request_filepath)
        cls._openssl('x509', '-req', '-sha256',
                     '-in', request_filepath, '-signkey', private_key_filepath, '-days', '36510',
                     '-extfile', extension_filepath, '-out', certificate_filepath)
        assert path.exists(certificate_filepath)

    @classmethod
    def generate_self_signed_certificate_for_web_server(cls):
        ssl = cls.get_webserver_ssl_configuration(False)
        cls._generate_certificate(ssl['certfile'], ssl['keyfile'])

    @classmethod
    def generate_self_signed_certificate_for_vcenter_extension(cls):
        ssl = cls.get_vcenter_extension_ssl_configuration()
        cls._generate_certificate(ssl['certfile'], ssl['keyfile'])

    @classmethod
    def generate_self_signed_certificates(cls):
        cls.generate_self_signed_certificate_for_vcenter_extension()
        cls.generate_self_signed_certificate_for_web_server()

    @classmethod
    def get_webapp_bind_address_and_port(cls):
        return (cls._get('address'), cls._get('plugin_http_port'))

    @classmethod
    def get_webapp_bind_address_and_ssl_port(cls):
        return (cls._get('address'), cls._get('plugin_https_port'))

    @classmethod
    def get_vasa_provider_bind_address_and_port(cls):
        return (cls._get('address'), cls._get('vasa_provider_https_port'))

    @classmethod
    def get_webserver_ssl_configuration(cls, attempt_to_write_certs_file=True):
        import ssl
        import os
        import shutil
        root_certificates = cls._get('root_certificates')
        my_certificate_path = cls._get('webserver_certificate').replace('\\', '')
        if attempt_to_write_certs_file and not os.path.exists(root_certificates):
            # wrap_socket fails if ca_certs file is empty or contains an invalid certificate
            # so we populate it with our own
            shutil.copy(my_certificate_path, root_certificates)
        return {'keyfile': cls._get('webserver_certificate_private_key').replace('\\', ''),
                'certfile': my_certificate_path,
                'cert_reqs': ssl.CERT_NONE,
                'ca_certs': root_certificates,
                'ssl_version': ssl.PROTOCOL_TLSv1}

    @classmethod
    def get_vcenter_extension_ssl_configuration(cls):
        return {'keyfile': cls._get('vcenter_extension_certificate_private_key').replace('\\', ''),
                'certfile': cls._get('vcenter_extension_certificate').replace('\\', '')}

    @classmethod
    def get_vcenter_address(cls):
        from infi.vendata.vmware_powertools.scripts.greeter import vCenterInvalidAddress
        from ..ovf import OVF
        ovf_attributes = OVF.get_attributes()
        if 'IP' not in ovf_attributes:
            raise vCenterInvalidAddress("vCenter IP not in OVF")
        if not ovf_attributes['IP']:
            raise vCenterInvalidAddress("vCenter IP is empry in OVF")
        return ovf_attributes['IP']

    @classmethod
    def _get_ip(cls, iface):
        from infi.vendata.vmware_powertools.utils import get_interface_ip_address
        return get_interface_ip_address(iface)

    @classmethod
    def _assert_ip_address(cls):
        from infi.vendata.vmware_powertools.scripts.greeter import NoIPAddress
        if not cls.is_in_production():
            return
        if cls._get_ip("eth0") is None:
            raise NoIPAddress("eth0")

    @classmethod
    def _get_source_address(cls):
        from ..ovf import OVF
        if cls.is_in_production() and not OVF.OVF_ENV.startswith("tests/ovf"):
            cls._assert_ip_address()
            return cls._get_ip("eth0")
        return cls._get_source_address_connected_to_vcenter()

    @classmethod
    def _get_source_address_connected_to_vcenter(cls):
        from socket import create_connection
        from infi.vendata.vmware_powertools.scripts.greeter import ConnectivityError
        vcenter_address = cls.get_vcenter_address()
        try:
            address = (vcenter_address, 80)
            connection = create_connection(address)
            source = connection.getsockname()
            connection.close()
            return source[0]
        except Exception, error:
            raise ConnectivityError(error)

    @classmethod
    def get_webapp_url(cls):
        _, port = cls.get_webapp_bind_address_and_port()
        return "{protocol}://{host}:{port}".format(protocol="http", host=cls._get_source_address(), port=port)

    @classmethod
    def get_webapp_ssl_url(cls):
        _, port = cls.get_webapp_bind_address_and_ssl_port()
        return "{protocol}://{host}:{port}".format(protocol="https", host=cls._get_source_address(), port=port)

    @classmethod
    def get_vasa_provider_url(cls):
        _, port = cls.get_vasa_provider_bind_address_and_port()
        return "{protocol}://{host}:{port}/vasa".format(protocol="https", host=cls._get_source_address(), port=port)

    @classmethod
    def get_vcenter_extension_key(cls):
        return cls._get('vcenter_extension_key')

    @classmethod
    def get_active_session_count(cls):
        return cls._get('active_session_count')

    @classmethod
    def get_rotating_logfile_attributes(cls):  # pragma: no cover
        return dict(max_size_in_bytes=cls._get('rotating_logfile_max_size') or DEFAULT_CONFIGURATION.get('rotating_logfile_max_size'),
                    backup_count=cls._get('rotating_logfile_backup_count') or DEFAULT_CONFIGURATION.get('rotating_logfile_backup_count'))

    @classmethod
    def get_vasa_provider_credentials(cls):
        return (cls._get('vasa_provider_username'), cls._get('vasa_provider_password'))
