import logging
import logbook
import random
import gevent
import greenlet
import gevent.local
import os
import re
from thread import get_ident as get_thread_id
from gevent.thread import get_ident as get_greenlet_id
from infi.pyutils.decorators import wraps
from infi.pyutils.contexts import contextmanager
from infi.vendata.vmware_powertools import DATA_DIR


logbook_application_setup = None
logger = logging.getLogger(__name__)
running_greenlets = {}
gevent_hub_id = id(gevent.get_hub())
LOG_DIRPATH = os.path.join(DATA_DIR, 'logs')
LOG_FILENAME = "server.log"
LOG_FILE_PATH = os.path.join(LOG_DIRPATH, LOG_FILENAME)
MODULU = 32768
MAX_DURATION = 1
TRUNCATION_LIMIT = 1024


def switch_time_tracer(what, greenlets):
    from time import time
    now = time()
    origin, target = greenlets
    running_greenlets[id(target)] = now, LoggingTag.get_tag()
    start_time, greenlet_tag = running_greenlets.pop(id(origin), (None, None))
    if greenlet_tag and start_time:
        duration = now-start_time
        if duration >= MAX_DURATION:
            msg = "greenlet id {} (tag {}) was running for {:.5f} seconds"
            logger.warn(msg.format(id(origin) % MODULU, greenlet_tag, duration))


def enable_logging_greenlet_switching():
    from time import time
    from os import getppid
    now = time()
    if not LoggingTag.get_tag():
        LoggingTag.set_random_tag()
    running_greenlets[id(gevent.getcurrent())] = now, LoggingTag.get_tag()
    msg = "enabling logging of greenlet switching, current greenlet (main) has tag {}, parent pid is {}"
    logger.debug(msg.format(LoggingTag.get_tag(), getppid()))
    greenlet.settrace(switch_time_tracer)


def disable_logging_greenlet_switching():
    greenlet.settrace(None)


class LoggingTag(object):
    _threadlocal = gevent.local.local()
    TAG_NAME = 'request_id'

    @classmethod
    def get_tag(cls):
        return getattr(cls._threadlocal, cls.TAG_NAME, None)

    @classmethod
    def set_tag(cls, tag):
        setattr(cls._threadlocal, cls.TAG_NAME, tag)

    @classmethod
    def set_random_tag(cls):
        cls.set_tag(cls.new_random_tag())

    @classmethod
    def new_random_tag(cls):
        return hex(random.getrandbits(32))[2:-1].zfill(8)


def _inject_tag(record):
    record.extra[LoggingTag.TAG_NAME] = LoggingTag.get_tag()


def _inject_info(record):
    if not record.extra.get('thread_id'):
        record.extra['thread_id'] = get_thread_id()
    if not record.extra.get('greenlet_id'):
        record.extra['greenlet_id'] = get_greenlet_id()


def file_formatter(record, handler):
    result = "{:%Y-%m-%d %H:%M:%S} ".format(record.time) + syslog_formatter(record, handler)
    return result


sub_password = lambda msg: re.sub('"password": "[^"]*"', '"password": "********"', msg)
sub_cherrypy_access = lambda msg: re.sub("vimSessionKey=[A-Za-z0-9_-]{36}", "vimSessionKey=********",
                                         re.sub("sessionId=[A-Za-z0-9_-]{36}", "sessionId=********", msg))
sub_traceback = lambda msg: re.sub("'sessionId': 'u?[A-Za-z0-9_-]{36}'", "'sessionId': '********'",
                                   re.sub("'vimSessionKey': u?'[A-Za-z0-9_-]{36}'", "'vimSessionKey': '********'", msg))

def syslog_formatter(record, handler):
    strformat = "pid={:0>5} tid={:0>5}:{:0>5} tag={:0>8} module={: <40} level={: <7} msg={}"
    greenlet_id = record.extra.get('greenlet_id', -1)
    thread_id = record.extra.get('thread_id', record.thread)
    tag = record.extra.get(LoggingTag.TAG_NAME, 0)
    trace = record.extra.get("trace", False)
    formatted_exception = record.formatted_exception
    exc_msg = '' if formatted_exception is None else '\n' + formatted_exception
    level = "TRACE" if trace else record.level_name
    message = record.message + exc_msg
    try:
        message = sub_traceback(sub_password(sub_cherrypy_access(message)))
    except:
        pass
    formatted = strformat.format(record.process,
                                 thread_id % MODULU, greenlet_id % MODULU,
                                 tag, record.channel, level, message)
    return formatted


def stderr_formatter(record, handler):
    import colorama
    res = file_formatter(record, handler)
    res = re.sub("(module=\S+)", colorama.Style.BRIGHT + colorama.Fore.RED + "\\1" + colorama.Style.RESET_ALL, res)
    res = re.sub("(level=\w+)", colorama.Style.BRIGHT + colorama.Fore.GREEN + "\\1" + colorama.Style.RESET_ALL, res)
    res = re.sub("(^\S+\s\S+)", colorama.Style.BRIGHT + colorama.Fore.BLUE + "\\1" + colorama.Style.RESET_ALL, res)
    return res


def new_handler(klass, formatter, *args, **kwargs):
    handler = klass(*args, **kwargs)
    handler.formatter = formatter
    return handler


@contextmanager
def empty_context():
    yield


def redirect_python_logging_to_logbook():
    # When we run nose tests we want to the opposite - redirect logbook to logging so nose will capture it.
    # So we shound't redirect back in this case otherwise we'll cause an infinite recursion loop.
    from logbook.compat import redirect_logging, LoggingHandler
    if not any(isinstance(obj, LoggingHandler) for obj in logbook.default_handler.stack_manager.iter_context_objects()):
        redirect_logging()


def clear_and_close_existing_handlers():
    global setup
    try:
        for item in setup.objects:      # noqa
            try:
                item.close()
            except:
                pass
    except:
        pass
    setup = None


class RotatingFileHandler(logbook.RotatingFileHandler):
    def perform_rollover(self):
        from logbook.handlers import rename, sys, errno
        self.stream.close()
        for x in xrange(self.backup_count - 1, 0, -1):
            src = '{}.{:02}'.format(self._filename, x)
            dst = '{}.{:02}'.format(self._filename, x + 1)
            try:
                rename(src, dst)
            except OSError:
                e = sys.exc_info()[1]
                if e.errno != errno.ENOENT:
                    raise
        rename(self._filename, self._filename + '.01')
        self._open('w')


def ensure_logging_directory_exists():
    import os
    # Ensure logging directory exists
    try:
        os.makedirs(LOG_DIRPATH)
    except OSError:
        if not os.path.isdir(LOG_DIRPATH):
            raise


def set_traceback_truncation_limit():
    from infi.traceback import set_truncation_limit
    set_truncation_limit(TRUNCATION_LIMIT)


def basic_config(to_file=True, to_stderr=False):  # pragma: no cover
    from infi.vendata.vmware_powertools.config import Configuration
    from logbook.concurrency import enable_gevent
    import sys
    enable_gevent()
    redirect_python_logging_to_logbook()
    clear_and_close_existing_handlers()
    processor = logbook.Processor(_inject_info)
    flags = logbook.Flags(errors='silent')
    handlers = [logbook.NullHandler(bubble=False)]
    logging_attributes = Configuration.get_rotating_logfile_attributes()
    if to_file:
        ensure_logging_directory_exists()
        handlers.append(new_handler(RotatingFileHandler, file_formatter,
                                    LOG_FILE_PATH, bubble=True, level=logbook.DEBUG,
                                    max_size=logging_attributes['max_size_in_bytes'],
                                    backup_count=logging_attributes['backup_count']))
    if to_stderr:
        handlers.append(new_handler(logbook.StreamHandler, stderr_formatter, sys.stderr,
                                    level=logbook.DEBUG, bubble=True))
    global logbook_application_setup
    setup = logbook.NestedSetup([processor, flags] + handlers)
    logbook_application_setup = setup.applicationbound()
    logbook_application_setup.__enter__()
    enable_logging_greenlet_switching()
    set_traceback_truncation_limit()


def logbook_tag(func, other_greenlet_tag=None):
    @wraps(func)
    def wrapped(*args, **kwargs):
        previous_tag = LoggingTag.get_tag()
        if previous_tag is None or other_greenlet_tag is not None:
            LoggingTag.set_random_tag() if other_greenlet_tag is None else LoggingTag.set_tag(other_greenlet_tag)
            if not hasattr(func, '_no_logging'):
                logger.debug("setting new tag {} on greenlet {}".format(LoggingTag.get_tag(), func.__name__))
        processor = logbook.Processor(_inject_tag)
        with processor.greenletbound():
            try:
                return func(*args, **kwargs)
            finally:
                LoggingTag.set_tag(previous_tag)
    return wrapped


@contextmanager
def log_timing_context(title, log_level="DEBUG"):
    from time import time
    t0 = time()
    log = getattr(logger, log_level.lower())
    log("started timing {} at {:.4f}".format(title, t0))
    try:
        yield
    finally:
        t1 = time()
        log("ended timing {} at {:.4f}. time taken: {:.4f} seconds".format(title, t1, t1 - t0))


def log_timing(func):
    @wraps(func)
    def wrapped(*args, **kwargs):
        with log_timing_context("function {!r}".format(func.func_name)):
            return func(*args, **kwargs)
    return wrapped
