"""infinivmware {version}
INFINIDAT Host Power Tools for VMware v{version}

Usage:
    infinivmware [options] [--production-environment] [--populate-storagemodel] [--with-vasa-provider] [--with-vwc] [--with-autoreload] [--signal-upstart] server
    infinivmware [options] shell
    infinivmware [options] client <vcenter-address> <username> <password> <esx-name>
    infinivmware [options] setup
    infinivmware [options] snapshot cleanup

Options:
    --ovf=<ovf>                        ovf filepath [default: /opt/vmware/etc/vami/ovfEnv.xml]
    --as-vcenter=VCENTER               connect to specific development server
    --vcenter-username=USERNAME        user for specific development server
    --vcenter-password=USERNAME        password for specific development server
    --as-mock
    --hosts=N                          number of ESX hosts [default: 6]
    --clusters=N                       number of ESX clusters [default: 6]
    --infiniboxes=N                    number of InfiniBoxes [default: 2]
    --initiator-ports=N                number of initiator ports [default: 3]
    --datastores=N                     number of datastores [default: 3]
    --virtual-machines=N               number of virtual-machines [default: 9]
    --map-to-lun0                      map volumes to LUN0
    --log-to-stderr
    --log-to-file

"""

from sys import argv
from infi.traceback import traceback_context
from infi.pyutils.contexts import contextmanager
from infi.pyutils.decorators import wraps
from logging import getLogger
logger = getLogger(__name__)

ERROR_LOGFILE_CONTENTS = """
Full log file is at http://{0}:5430/
Scroll down for the complete log.

Short {1}

{2}
"""

active_traceback_context = None


def traceback_decorator(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        global active_traceback_context
        assert active_traceback_context is None
        active_traceback_context = traceback_context()
        active_traceback_context.__enter__()
        try:
            return func(*args, **kwargs)
        finally:
            active_traceback_context.__exit__(None, None, None)
            active_traceback_context = None
    return wrapper


@contextmanager
def disable_traceback_context():
    global active_traceback_context
    assert active_traceback_context is not None
    active_traceback_context.__exit__(None, None, None)
    active_traceback_context = None
    try:
        yield
    finally:
        active_traceback_context = traceback_context()
        active_traceback_context.__enter__()


def _setup():
    from .upstart import install, restart_after
    from logging import basicConfig, DEBUG
    basicConfig(level=DEBUG, file='/var/log/vmware_powertools/post_install.log', filemode='a')
    with restart_after():
        install()

@contextmanager
def _empty_context(**kwargs):
    yield


def _turn_on_upstart_debug_logging():
    from os import path
    from infi.execute import execute
    INITCTL = "/sbin/initctl"
    if path.exists(INITCTL):
        execute([INITCTL, "log-priority", "debug"])


def _write_error_logfile():
    from traceback import format_exc
    from infi.vendata.vmware_powertools.scripts.greeter import ERROR_LOGFILE
    from infi.vendata.vmware_powertools.utils import get_interface_ip_address
    ip_address = get_interface_ip_address("eth0") or "<virtual-machine-ip-address>"
    with disable_traceback_context():
        short_tracback = "\n".join(format_exc().splitlines()[::2])
    long_traceback = format_exc()

    with open(ERROR_LOGFILE, "w") as fd:
        fd.write(ERROR_LOGFILE_CONTENTS.format(ip_address, short_tracback, long_traceback))


def _server(production_environment, populate_storagemodel,
            with_vasa_provider, with_vwc,
            with_autoreload, signal_upstart):
    from .starter import _start_all, _stop_all
    from .upstart import signal_init_that_i_am_ready
    from ..webserver import WebServer
    from .autoreload import autoreload
    from infi.vendata.vmware_powertools.scripts.greeter import rotate_error_logfile, PrettyError, ERROR_LOGFILE
    from gevent import sleep
    rotate_error_logfile()
    _turn_on_upstart_debug_logging()
    try:
        _start_all(production_environment, populate_storagemodel, with_vasa_provider, with_vwc)
        server_func = autoreload(WebServer.serve_forever) if with_autoreload else WebServer.serve_forever
    except PrettyError, error:
        logger.exception("error starting up the server on pretty error, lets sleep a bit")
        with open(ERROR_LOGFILE, "w") as f:
            f.write(str(error))
        _stop_all(with_vasa_provider=True)
        sleep(10)
        raise
    except:
        logger.exception("error starting up the server")
        _write_error_logfile()
        _stop_all(with_vasa_provider=True)
        raise
    finally:  # Due to upstart bug #712351, we must signal upstart even if we're about to die.
        if signal_upstart:
            signal_init_that_i_am_ready()
    with open(ERROR_LOGFILE, "w") as f:
        pass
    try:
        server_func()
        _stop_all(with_vasa_provider=True)
    except:
        logger.exception("error starting up the server")
        raise

def _shell():
    from ..vcenter import vCenter
    from IPython import embed
    from infi.storagemodel import get_storage_model       # noqa
    from pyVmomi import vim       # noqa
    vCenter.register_extension_through_vservice()
    vCenter.prepare_property_collectors()
    client = vCenter.get_vcenter_client()       # noqa
    embed()

def _client(vcenter_address, username, password, esx_name):
    from IPython import embed
    from infi.pyvmomi_wrapper import Client
    from infi.storagemodel.vmware import with_host
    client = Client(vcenter_address, username, password)
    esx = client.get_host_system(esx_name)
    context = with_host(client, esx)
    context.__enter__()
    embed()

@contextmanager
def _prepare_context(arguments):
    from gevent import monkey
    from infi.vendata.vmware_powertools.patches import powertools_patches
    from ..logging import basic_config
    from ..vmock import vMockFactory
    from ..vmock.as_vcenter import vCenterLoginPatcher
    from ..ovf import OVF
    monkey.patch_all()
    powertools_patches.patch_all()
    basic_config(arguments.get('--log-to-file'), arguments.get('--log-to-stderr'))
    original_ovf = OVF.OVF_ENV
    OVF.OVF_ENV = arguments.get('--ovf')
    context_func = _empty_context
    context_kwargs = dict()
    if arguments.get("--as-mock"):
        context_kwargs = dict(
            hosts=int(arguments.get("--hosts", 0)),
            clusters=int(arguments.get("--clusters", 0)),
            infiniboxes=int(arguments.get("--infiniboxes", 0)),
            initiator_ports=int(arguments.get("--initiator-ports", 0)),
            datastores=int(arguments.get("--datastores", 0)),
            virtual_machines=int(arguments.get("--virtual-machines", 0)),
            map_to_lun0=arguments.get('--map-to-lun0', False),
        )
        context_func = vMockFactory.create
    elif arguments.get("--as-vcenter"):
        context_kwargs = dict(vcenter_address=arguments.get("--as-vcenter"),
                              username=arguments.get("--vcenter-username"),
                              password=arguments.get("--vcenter-password"))
        context_func = vCenterLoginPatcher
    try:
        with context_func(**context_kwargs):
            yield
    finally:
        OVF.OVF_ENV = original_ovf


@traceback_decorator
def main(argv=argv[1:]):
    from docopt import docopt
    from infi.vendata.vmware_powertools.__version__ import __version__
    import requests
    requests.packages.urllib3.disable_warnings()

    arguments = docopt(__doc__.format(version=__version__), version=__version__, argv=argv)
    if arguments.get("setup"):
        return _setup()
    with _prepare_context(arguments):
        if arguments.get("server"):
            return _server(arguments['--production-environment'],
                           arguments['--populate-storagemodel'],
                           arguments['--with-vasa-provider'],
                           arguments['--with-vwc'],
                           arguments['--with-autoreload'],
                           arguments['--signal-upstart'])
        elif arguments.get("shell"):
            return _shell()
        elif arguments.get("client"):
            return _client(arguments['<vcenter-address>'], arguments['<username>'], arguments['<password>'],
                           arguments['<esx-name>'])
        elif arguments.get("snapshot") and arguments.get("cleanup"):
            from infi.vendata.vmware_powertools.snapshots.cleanup import cleanup_stale_snapshots
            return cleanup_stale_snapshots()
