from logging import getLogger

logger = getLogger(__name__)

def _populate_storagemodel(client=None):
    from ...patches import with_host
    from ...vcenter import vCenter
    client = client or vCenter.get_extension_client()
    logger.info("Starting storagemodel population")
    for host in client.get_host_systems():
        with with_host(host):
            from infi.storagemodel.vendor.infinidat.shortcuts import get_infinidat_block_devices
            from infi.storagemodel.vendor.infinidat.shortcuts import get_infinidat_storage_controller_devices
            from infi.storagemodel import get_storage_model
            get_storage_model().refresh()
            for device in get_infinidat_block_devices() + get_infinidat_storage_controller_devices():
                _ = device.get_vendor()         # noqa (caching)
    logger.info("Finished storagemodel population")

def _disable_caching_for_staticdir():
    from infi.pyutils.decorators import wraps
    import cherrypy
    import mock
    patcher = mock.patch.object(cherrypy.lib.static.cptools, 'validate_since')
    def patch_calls_to_validate_since(func):
        @wraps(func)
        def callable(*args, **kwargs):
            with patcher as patch:
                patch.return_value = False
                return func(*args, **kwargs)
        return callable
    cherrypy.lib.static.staticdir = patch_calls_to_validate_since(cherrypy.lib.static.staticdir)

def _start_all(production_environment, populate_storagemodel, with_vasa_provider, with_vwc):  # pragma: no cover
    from ...utils import auth  # noqa - for replaceing the CREDENTIAL_STORE singleton
    from ...redis import RedisServer
    from ...webserver import WebServer
    from ...config import Configuration
    from ...vcenter import vCenter
    from ...vcenter.stale_tasks import cancel_stale_tasks
    from ...webapp import WebApplication
    from ...webapp.plugin.resource import set_ready
    from ...vwc import register_vwc
    _disable_caching_for_staticdir()
    RedisServer.start()
    Configuration.populate_defaults(production_environment)
    WebServer.start(with_vasa_provider)
    vCenter.register_extension_through_vservice()
    vCenter.update_vm_version_if_necessary()
    vCenter.prepare_property_collectors()
    WebApplication.clear_cache()
    try:
        cancel_stale_tasks()
    except:
        logger.exception("failed to cancel stale tasks")
    set_ready()
    if populate_storagemodel:
        _populate_storagemodel()
    if with_vasa_provider:
        vCenter.register_vasa_provider()
    if with_vwc:
        register_vwc()

def _stop_all(with_vasa_provider=True):  # pragma: no cover
    from ...redis import RedisServer
    from ...webserver import WebServer
    from ...vcenter import vCenter
    from ...webapp import WebApplication
    from infi.storagemodel.vmware.patches.storagemodel import StorageModelFactory
    vCenter.stop_keepalive_greenlet()
    WebServer.stop(with_vasa_provider)
    RedisServer.stop()
    try:
        vCenter.get_extension_client().logout()
    except:
        pass
    vCenter.clear_cache_for_get_client_methods()
    StorageModelFactory.clear()
    RedisServer.clear_cache()
    try:
        WebApplication.clear_cache()
    except:
        pass

def _install_signal_handlers():  # pragma: no cover
    import signal
    import os
    if os.name == 'nt':
        return
    signal.siginterrupt(signal.SIGCHLD, False)
    signal.siginterrupt(signal.SIGALRM, False)
