from infi.pyutils.contexts import contextmanager
from infi.vendata.vmware_powertools.vcenter import vCenter
from infi.vendata.vmware_powertools.vcenter.shortcuts import wait_for_task
from . import exceptions
from .snapshotstore import SnapshotStore
from pyVmomi import vim
import uuid
import logging

logger = logging.getLogger(__name__)


def find_new_name(client, snapshot_info):
    VSPHERE_MAX_CHARS = 80
    SUFFIX_LENGTH = len("-revert000")
    original_vm_name = snapshot_info["vm_name"][:(VSPHERE_MAX_CHARS - SUFFIX_LENGTH)]
    count = 0
    while count < 1000:
        new_name = "{}-revert{}".format(original_vm_name, "" if count == 0 else count)
        if client.get_virtual_machine(new_name) is None:
            return new_name
        count += 1
    # if we got here that means that all 1000 snapshot names are taken!
    logger.info("Couldn't find new name for reverted VM. User will have to choose...")
    return ""


def get_default_revert_params(client, vm, snapshot_info):
    from infi.vendata.vmware_powertools.vcenter.shortcuts import get_host_by_uuid, get_datastore_by_url
    logger.debug("original host uuid is {}. original datastore url is {}. Folder moref is {}".format(
        snapshot_info["original_host_uuid"],
        snapshot_info["original_datastore_url"],
        snapshot_info["original_folder_ref"]))
    host = get_host_by_uuid(client, snapshot_info["original_host_uuid"])
    if host is None:
        if vm is not None:
            host = vm.summary.runtime.host
            logger.info("Using vm's current host, uuid {}".format(host.summary.hardware.uuid))
        else:
            logger.info("Original host not found. User will have to choose...")
    datastore = get_datastore_by_url(client, snapshot_info["original_datastore_url"])
    if datastore is None:
        logger.info("Original datastore not found. Trying to choose datastore from host")
        if host is not None:
            datastore = host.datastore[0]
        else:
            logger.info("User will have to choose datatore...")
    try:
        folder = client.get_managed_object_by_reference(snapshot_info["original_folder_ref"])
    except ValueError:
        logger.info("Original folder not found. Using host's datacenter's main folder")
        if host is not None:
            datacenter = host.parent.parent.parent
            folder = datacenter.vmFolder
        else:
            folder = None
            logger.info("User will have to choose folder...")
    name = find_new_name(client, snapshot_info)
    logger.info("New suggested name is {}".format(name))
    return name, host, datastore, folder


def wait_for_volume(host, volume):
    from infi.vendata.vmware_powertools.patches import with_host
    with with_host(host):
        from infi.storagemodel.vendor.infinidat.predicates import InfinidatVolumeExists
        from infi.storagemodel import get_storage_model
        model = get_storage_model()
        predicate = InfinidatVolumeExists(volume.get_system().get_serial(), volume.get_id())
        model.rescan_and_wait_for(predicate)


def find_volume_device(host, volume):
    from infi.vendata.vmware_powertools.patches import with_host
    with with_host(host):
        from infi.storagemodel.vendor.infinidat.shortcuts import get_infinidat_block_devices
        for dev in get_infinidat_block_devices():
            vendor = dev.get_vendor()
            volume_id = volume.get_id()
            system_serial = volume.get_system().get_serial()
            if vendor.get_volume_id() == volume_id \
               and vendor.get_system_serial() == system_serial:
                return dev
    raise exceptions.VolumeNotMapped("Could not find device mapped for volume id {}".format(volume.get_id()))


def mount_datastore_snapshot(host, volume_device):
    device_path = volume_device.get_block_access_path() + ":1"
    resignature_spec = vim.HostUnresolvedVmfsResignatureSpec(extentDevicePath=device_path)
    task = host.configManager.datastoreSystem.ResignatureUnresolvedVmfsVolume_Task(resignature_spec)
    wait_for_task(task)
    result = task.info.result.result
    return result


class RevertSnapshotOperation(object):
    def __init__(self, requesting_object, vm_uuid, snapshot_id, resignature_vm, vm_name, host, datastore, folder):
        self._requesting_object = requesting_object
        self._should_resignature_vm = resignature_vm
        self._snapshot_info = SnapshotStore.get_snapshot_info(SnapshotStore.VM_SNAP_STORE, vm_uuid, snapshot_id)
        self._update_call_count = 0
        self._vm_name = vm_name
        self._host = host
        self._datastore = datastore
        self._folder = folder

    @contextmanager
    def _map_snapshot(self, host, clean=True):
        from infinisdk.core.exceptions import ObjectNotFound
        from infi.vendata.vmware_powertools.patches import with_host
        from infi.vendata.vmware_powertools.utils import get_system_object_and_host
        self._update_task_description("Mapping snapshot volume to host {}".format(host.name))
        infinibox_serial = self._snapshot_info["infinibox_serial"]
        with with_host(host):
            infinisdk_system, infinisdk_host = get_system_object_and_host(infinibox_serial)
        infinibox_snapshot_id = self._snapshot_info['infinibox_snapshot_id']
        logger.debug("Infinibox Snapshot id={}".format(infinibox_snapshot_id))
        try:
            snapshot_vol = infinisdk_system.volumes.get(id=infinibox_snapshot_id)
        except ObjectNotFound:
            msg = "Can't find snapshot on system {} (volume id {}). Possibly it was deleted."
            raise exceptions.SnapshotNotFound(msg.format(infinibox_serial, infinibox_snapshot_id))
        cloned_vol = snapshot_vol.create_clone()
        infinisdk_host.map_volume(cloned_vol)
        try:
            yield cloned_vol
        finally:
            if clean:
                self._update_task_description("Cleaning volume")
                infinisdk_host.unmap_volume(cloned_vol)
                cloned_vol.delete()
                self._rescan_host(host)

    def _rescan_host(self, host):
        from infi.vendata.vmware_powertools.patches import with_host
        with with_host(host):
            from infi.storagemodel import get_storage_model
            model = get_storage_model()
            self._update_task_description("Rescanning host")
            model.rescan_and_wait_for()

    @contextmanager
    def _find_new_datastore(self, host, cloned_vol, clean=True):
        self._update_task_description("Rescanning host")
        wait_for_volume(host, cloned_vol)
        self._update_task_description("Locating device")
        volume_device = find_volume_device(host, cloned_vol)
        self._update_task_description("Resolving new VMFS volume")
        new_datastore = mount_datastore_snapshot(host, volume_device)
        try:
            yield new_datastore
        finally:
            if clean:
                self._update_task_description("Cleaning snapshot datastore")
                host.configManager.storageSystem.UnmountVmfsVolume(new_datastore.info.vmfs.uuid)
                try:
                    host.configManager.storageSystem.DetachScsiLun(volume_device.get_uuid())
                except vim.HostConfigFault:
                    logger.error("Got HostConfigFault on DetachScsiLun. See HIPVM-601")

    def _register_vm(self, new_datastore, host):
        self._update_task_description("Registering Virtual Machine")
        datastore_path = self._snapshot_info["vm_datastore_path"].format(new_datastore.name)
        cluster = host.parent
        if not self._vm_name:
            raise exceptions.NameNotAvailable("Empty name is not allowed.")
        if self._client.get_virtual_machine(self._vm_name):
            raise exceptions.NameNotAvailable("The chosen name is already in use.")
        task = self._folder.RegisterVM_Task(datastore_path,
                                            name=self._vm_name,
                                            asTemplate=False,
                                            pool=cluster.resourcePool,
                                            host=host)
        wait_for_task(task)
        return task.info.result

    def _relocate_vm(self, vm, datastore):
        import time
        self._update_task_description("Relocating Virtual Machine")
        relocate_spec = vim.VirtualMachineRelocateSpec(datastore=datastore,
                                                       transform=None,
                                                       host=None,
                                                       pool=None)
        task = vm.RelocateVM_Task(relocate_spec, None)
        wait_for_task(task)
        # For some weird reason, we can't delete the snapshot immediately after relocation, so sleep a bit...
        time.sleep(5)

    def _update_task_description(self, new_description):
        logger.debug("Task description: {}".format(new_description))
        self._task.set_description(new_description)
        percentage = int((self._update_call_count) * (100.0 / self._operation_count))
        self._update_call_count += 1
        self._task.update_progress(percentage)

    def _resignature_vm(self, vm):
        self._update_task_description("Resignaturing Virtual Machine")
        cfg_spec = vim.VirtualMachineConfigSpec(uuid=str(uuid.uuid1()))
        task = vm.ReconfigVM_Task(cfg_spec)
        wait_for_task(task)

    def _unregister_vm(self, vm):
        vm.UnregisterVM()

    def revert(self, expose_only=False):
        from infi.vendata.vmware_powertools.vcenter.shortcuts import create_vcenter_task
        logger.debug(("exposing" if expose_only else "restoring") + " snapshot")
        self._client = vCenter.get_extension_client()
        task_name = ("Expose" if expose_only else "Restore") + " INFINIDAT Snapshot"
        self._task = create_vcenter_task(self._requesting_object, task_name)
        self._operation_count = 5           # map, rescan, locate, resolve, register
        if not expose_only:
            self._operation_count += 4      # relocate, clean ds, clean vol, rescan
        if self._should_resignature_vm:
            self._operation_count += 1      # resig

        with self._task:

            # make sure datastore is visible to host
            logger.debug("uuid of host to be used is {}".format(self._host.summary.hardware.uuid))
            if not expose_only:
                # datastore is not important if we expose_only, but if we need to relocate, then
                # the datastore details are important and the check is needed
                host_datastore_refs = [datastore._moId for datastore in self._host.datastore]
                logger.debug("all datastores in target host: {}".format(host_datastore_refs))
                logger.debug("url of datastore to be relocate to: {}".format(self._datastore._moId))
                if self._datastore._moId not in host_datastore_refs:
                    raise exceptions.DatastoreNotAvailable("Target datastore is not visible to target host")

            with self._map_snapshot(self._host, clean=not expose_only) as cloned_vol:
                with self._find_new_datastore(self._host, cloned_vol, clean=not expose_only) as new_datastore:
                    new_vm = self._register_vm(new_datastore, self._host)
                    if not expose_only:
                        try:
                            self._relocate_vm(new_vm, self._datastore)
                        except:
                            self._unregister_vm(new_vm)
                            raise
                    if self._should_resignature_vm:
                        self._resignature_vm(new_vm)

        if expose_only:
            return self._vm_name, new_datastore.name, cloned_vol.get_name()
        else:
            return self._vm_name, None, None
