from pyVmomi import vim


class VendorQuery(object):
    @classmethod
    def _get_datastore_device(cls, disk_name, esx_host):
        from infi.vendata.vmware_powertools.patches import with_host
        with with_host(esx_host):
            from infi.storagemodel import get_storage_model
            model = get_storage_model()
            multipath = model.get_native_multipath()
            multipath_devices = multipath.get_all_multipath_block_devices()
            for device in multipath_devices:
                if device.get_canonical_name() == disk_name:
                    return device
        # failed to get it from the cache, lets try again after clearing it
        with with_host(esx_host):
            from infi.storagemodel import get_storage_model
            model = get_storage_model()
            model.refresh()
            multipath = model.get_native_multipath()
            multipath_devices = multipath.get_all_multipath_block_devices()
            for device in multipath_devices:
                if device.get_canonical_name() == disk_name:
                    return device
        # bummer
        return None

    @classmethod
    def get_datastore_infinidat_vendor(cls, datastore):
        try:
            assert datastore.summary.type == "VMFS"
            assert len(datastore.info.vmfs.extent) == 1
            disk_name = datastore.info.vmfs.extent[0].diskName
            assert len(datastore.host) > 0

            device = cls._get_datastore_device(disk_name, datastore.host[0].key)
            assert device is not None
            assert device.get_scsi_vendor_id() == "NFINIDAT"
            return device.get_vendor()
        except AssertionError:
            return None

    @classmethod
    def get_vm_infinidat_vendor(cls, vm_object):
        if len(cls.vm_datastores(vm_object)) != 1:
            return None, None
        datastore = cls.vm_datastores(vm_object)[0]
        vendor = cls.get_datastore_infinidat_vendor(datastore)
        return datastore, vendor

    @classmethod
    def get_system_serial_from_datastore(cls, vm_object):
        datastore, vendor = cls.get_vm_infinidat_vendor(vm_object)
        if vendor is None:
            return None
        return vendor.get_system_serial()

    @classmethod
    def get_system_name_and_serial_from_datastore(cls, vm_object):
        datastore, vendor = cls.get_vm_infinidat_vendor(vm_object)
        if vendor is None:
            return None
        return vendor.get_system_name(), vendor.get_system_serial()

    # functions to determine what we should show about a VM in vm/folder/datastore views.
    # Each VM will be classified as one of the following:
    # 1. no infinidat datastore
    # 2. one infinidat datastore but more datastores/rdm volumes available (won't be available for snapshot)
    # 3. exactly one infinidat datastore (snapshot available)

    @classmethod
    def vm_datastores(cls, vm_object):
        return list(set(
                dev.backing.datastore for dev in vm_object.config.hardware.device if isinstance(dev, vim.VirtualDisk)
                and isinstance(dev.backing, vim.VirtualDiskFlatVer2BackingInfo)))

    @classmethod
    def is_too_many_datastores(cls, vm_object):
        def is_rdm(device):
            return isinstance(device, vim.VirtualDisk) and \
                   isinstance(device.backing, vim.VirtualDiskRawDiskMappingVer1BackingInfo)
        has_rdm_disk = any(is_rdm(device) for device in vm_object.config.hardware.device)
        return len(cls.vm_datastores(vm_object)) != 1 or has_rdm_disk

    @classmethod
    def has_infinidat_datastore(cls, vm_object):
        return any(cls.get_datastore_infinidat_vendor(datastore) is not None
                   for datastore in cls.vm_datastores(vm_object))

    @classmethod
    def one_infinidat_datastore_and_more(cls, vm_object):
        return cls.has_infinidat_datastore(vm_object) and cls.is_too_many_datastores(vm_object)

    # if has_infinidat_datastore is True and one_infinidat_datastore_and_more is False, it means
    # that get_vm_infinidat_vendor will not return None
