from infi.vendata.powertools.utils import queries
from infi.vendata.vmware_powertools.__version__ import __version__
from logging import getLogger


logger = getLogger(__name__)


def update_system_object(system_object):
    system_object.api.set_source_identifier("VMPowerTools/{}".format(__version__))

def get_system_object_by_device_vendor(vendor, log_in=True):
    result = queries.get_system_object_by_device_vendor(vendor, log_in)
    update_system_object(result)
    return result

def get_system_object_by_device(device, log_in=True):
    result = queries.get_system_object_by_device(device, log_in)
    update_system_object(result)
    return result

def get_system_object_by_serial(system_serial, log_in=True):
    result = queries.get_system_object_by_serial(system_serial, log_in)
    update_system_object(result)
    return result

def get_system_object_and_host(system_serial, log_in=True):
    system, host = queries.get_system_object_and_host(system_serial, log_in)
    update_system_object(system)
    return system, host

def get_infinisdk_volume_by_device(infinidat_device):
    system = get_system_object_by_device(infinidat_device)
    vendor = infinidat_device.get_vendor()
    return system.volumes.get(id=vendor.get_volume_id())

def get_infinisdk_host_and_volume_by_device(infinidat_device):
    system = get_system_object_by_device(infinidat_device)
    vendor = infinidat_device.get_vendor()
    volume = system.volumes.get(id=vendor.get_volume_id())
    host = system.hosts.get(id=vendor.get_host_id())
    return host, volume

def get_infinidat_device_by_serial__inner(volume_serial):
    from infi.storagemodel.vendor.infinidat.shortcuts import get_infinidat_block_devices
    serials = set()
    for device in get_infinidat_block_devices():
        if volume_serial == device.get_scsi_serial_number():
            return device
        else:
            serials.add(device.get_scsi_serial_number())
    logger.debug("no device matching scsi serial {}, found others:{} ".format(volume_serial, serials))
    return None

def get_infinidat_device_by_serial(esx_host, volume_serial):
    from infi.vendata.vmware_powertools.patches import with_host
    with with_host(esx_host):
        return get_infinidat_device_by_serial__inner(volume_serial)

def spawn_later(seconds, func, *args, **kwargs):
    """
    this function eventually calls gevent.spawn_later,
    but it also copies greenlet-local storage data we're using across the project
    """
    import cherrypy
    import gevent

    def _wrapper(local_dict):
        copy_local_dicts = hasattr(cherrypy.serving, '_local__dicts')
        if copy_local_dicts:
            cherrypy.serving._local__dicts[gevent.getcurrent()] = dict(local_dict)
        try:
            return func(*args, **kwargs)
        except Exception:
            logger.exception("caught exception in greenlet {!r} {!r} {!r}".format(func, args, kwargs))
            raise
        finally:
            if copy_local_dicts:
                cherrypy.serving._local__dicts.pop(gevent.getcurrent(), None)

    copy_local_dicts = hasattr(cherrypy.serving, '_local__dicts')
    local_dict = cherrypy.serving._local__dicts[gevent.getcurrent()] if copy_local_dicts else {}
    return gevent.spawn_later(seconds, _wrapper, local_dict)

def spawn(func, *args, **kwargs):
    return spawn_later(0, func, *args, **kwargs)

import gevent
joinall = gevent.joinall


def refresh_storage_model_on_host(esx_host, predicate=None, wait_for_completion=True, raise_error=False, do_rescan=True, do_refresh=False):
    """
    Rescan the given host and optionally wait for the given predicate.
    """
    from infi.vendata.vmware_powertools.patches import with_host
    with with_host(esx_host):
        from infi.storagemodel import get_storage_model
        get_storage_model().rescan_and_wait_for(predicate, wait_for_completion=wait_for_completion, raise_error=raise_error,
                                                do_rescan=do_rescan, do_refresh=do_refresh)


def refresh_storage_model_on_cluster(esx_cluster, predicate=None, wait_for_completion=True, raise_error=False, do_rescan=True, do_refresh=False):
    """
    Rescan all hosts in the given cluster (in parallel) and optionally wait for the given predicate.
    """
    from infi.vendata.vmware_powertools.mapping import good_cluster_hosts
    joinall([spawn(refresh_storage_model_on_host, esx_host, predicate, wait_for_completion, raise_error, do_rescan, do_refresh) for
             esx_host in good_cluster_hosts(esx_cluster)], raise_error=raise_error)


def get_interface_ip_address(iface):
    import socket
    import struct
    import fcntl
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sockfd = sock.fileno()
    SIOCGIFADDR = 0x8915
    ifreq = struct.pack('16sH14s', iface, socket.AF_INET, '\x00'*14)
    try:
        res = fcntl.ioctl(sockfd, SIOCGIFADDR, ifreq)
    except:
        return None
    ip = struct.unpack('16sH2x4s8x', res)[2]
    return socket.inet_ntoa(ip)


def find_and_kill_listening_process(port):
    from psutil import process_iter, AccessDenied

    def _listens_on_port(process):
        try:
            for connection in process.connections():
                try:
                    if connection.laddr[1] == port and connection.status == 'LISTEN':
                        return True
                except:
                    logger.exception('bad connection')

        except AccessDenied:
            pass
        except:
            logger.exception('bad process')

    def _kill(process):
        logger.debug("killing {!r}".format(process))
        try:
            process.kill()
        except:
            logger.exception("failed to kill {!r}".format(process))
            raise
        logger.debug("{!r} was killed".format(process))


    processes = list(process_iter())
    logger.debug("iterating over {} processes, looking for those who liste on {}".format(len(processes), port))
    processes_to_kill = [process for process in processes if _listens_on_port(process)]
    logger.debug("need to kill {} processes".format(len(processes_to_kill)))
    for process in processes_to_kill:
        _kill(process)
