import time
from ZSI import FaultException
from .zsi.vasaService_types import ns2
from .entities import Entities, iterate_hosts_with_callback
from .utils import client_sessions, decorators
from logging import getLogger
from infi.storagemodel.vendor.infinidat.infinibox.naa import InfinidatNAA
logger = getLogger(__name__)

# we must not return more than 100 events in a single request. described in the VASA spec
MAX_EVENTS = 100
CONFIG_MAP = {"HOST_MAPPED": "New", "HOST_UNMAPPED": "Delete", "VOLUME_RENAMED": "Update", "VOLUME_RESIZED": "Update"}

class VasaProviderEventsMixin(object):
    """ implements getEvents """

    def __init__(self):
        # dict of session-id -> (dict of system serial-> last event id seen from that system)
        self.session_to_last_event_id = dict()
        # dict of session-id -> next event id for the session
        self.session_to_event_id = dict()
        # dict of session-id -> list of events generated during the session (truncated after every request)
        self.session_to_events = dict()

    def _init_event_session(self, session_id):
        self.session_to_event_id[session_id] = 0
        self.session_to_events[session_id] = list()
        last_event_id_dict = dict()
        for serial, infinisdk in Entities.get_infinisdk_dict().items():
            self._init_last_id(serial, infinisdk, last_event_id_dict)
        self.session_to_last_event_id[session_id] = last_event_id_dict

    def _deinit_event_session(self, session_id):
        self.session_to_event_id.pop(session_id, None)
        self.session_to_events.pop(session_id, None)
        self.session_to_last_event_id.pop(session_id, None)

    def _refresh_storage_system(self):
        def callback():
            from infi.storagemodel import get_storage_model
            get_storage_model().refresh()
            get_storage_model().rescan_and_wait_for()
            yield None              # callback must be an iterator
        list(iterate_hosts_with_callback(callback))     # we convert to list of Nones so that the iterator will be consumed

    def _next_event_id(self):
        session_id = client_sessions.get_session().get_session_id()
        next_id = self.session_to_event_id.setdefault(session_id, 0)
        self.session_to_event_id[session_id] += 1
        return next_id

    def _generate_volume_event(self, system_serial, volume_id, config_type):
        new_event = self.local.response.new_return()
        new_event.set_element_eventId(self._next_event_id())
        new_event.set_element_eventType("Config")
        new_event.set_element_eventObjType("StorageLun")
        new_event.set_element_eventConfigType(config_type)
        new_event.set_element_eventTimeStamp(time.localtime())
        # the volume serial is not in the event, and we can't find the volume (it may be delted) so
        # we build the volume id (naa) according to the infinidat serial format
        volume_naa = InfinidatNAA.from_volume_id_and_system_serial(volume_id, system_serial)
        new_event.set_element_objectId(str(volume_naa))
        return new_event

    def _event_from_system_event(self, system_event, system_serial):
        volume_id = [data["value"] for data in system_event["data"] if data["name"] == "volume_id"][0]
        msg = "getEvents - found relevant event. code={} volume_id={}"
        logger.debug(msg.format(system_event["code"], volume_id))
        config_type = CONFIG_MAP[system_event["code"]]
        return self._generate_volume_event(system_serial, int(volume_id), config_type)

    def _init_last_id(self, system_serial, system_infinisdk, last_event_id_dict):
        if system_serial in last_event_id_dict:
            return
        last_event_id_dict[system_serial] = system_infinisdk.events.get_last_event()["id"]
        logger.debug("first time querying events from system {}. "
                     "Remembering {} as last event in system".format(system_serial, last_event_id_dict[system_serial]))

    def _get_events_from_infinisdk(self, system_infinisdk, last_event_id_dict):
        system_serial = system_infinisdk.get_serial()
        self._init_last_id(system_serial, system_infinisdk, last_event_id_dict)
        last_id_seen_from_system = last_event_id_dict[system_serial]
        msg = "getEvents - getting system events from serial {} - from id {} forward"
        logger.debug(msg.format(system_serial, last_id_seen_from_system))
        for system_event in system_infinisdk.events.get_events(last_id_seen_from_system+1):
            if system_event["code"] in CONFIG_MAP.keys():
                yield self._event_from_system_event(system_event, system_serial)
            last_event_id_dict[system_serial] = system_event["id"]

    @decorators.validate_certificate
    @decorators.validate_session
    @decorators.soap_decorator
    def soap_getEvents(self):
        session_id = client_sessions.get_session().get_session_id()
        next_id = self.session_to_event_id.setdefault(session_id, 0)

        last_id = self.local.request.get_element_lastEventId()
        if last_id != -1 and last_id < next_id - 1:
            raise FaultException(ns2.InvalidArgument_Dec().pyclass())

        last_event_id_dict = self.session_to_last_event_id.setdefault(session_id, dict())
        events = self.session_to_events.setdefault(session_id, list())
        for serial, infinisdk in Entities.get_infinisdk_dict().items():
            events.extend(self._get_events_from_infinisdk(infinisdk, last_event_id_dict))
        if len(events) > 0:
            # lun mappings changed, so rescan. we do it here because we don't want to rescan for every request,
            # and this is the only place the certification checks these sort of changes
            self._refresh_storage_system()
        logger.debug("getEvents - {} events to return".format(len(events)))
        self.local.response.set_element_return(events[:MAX_EVENTS])
        self.session_to_events[session_id] = events[MAX_EVENTS:]
