from logging import getLogger
logger = getLogger(__name__)


class HIPVM_251_WORKAROUND(object):
    _address_to_cert = dict()

    @classmethod
    def get(cls, address):
        for key, value in cls._address_to_cert.items():
            if address == key:
                return value

    @classmethod
    def add(cls, address, certificate_id):
        cls._address_to_cert[address] = certificate_id

    @classmethod
    def remove(cls, address):
        if address in cls._address_to_cert:
            cls._address_to_cert.pop(address)


class Certificate(object):
    def __init__(self, pem_certificate, certificate):
        """
        :param pem_certificate: returned by ssl.DER_cert_to_PEM_cert(SSLSocket.getpeercert(True))
        :param certificate: returned by SSLSocket.getpeercert(False)
        """
        super(Certificate, self).__init__()
        self._pem_certificate = pem_certificate
        self._certificate = certificate

    @classmethod
    def not_after_string_to_datetime_object(cls, not_after):
        from datetime import datetime
        from ssl import cert_time_to_seconds
        return datetime.utcfromtimestamp(cert_time_to_seconds(not_after))

    def is_expired(self):
        from datetime import datetime
        return self.not_after_string_to_datetime_object(self._certificate['notAfter']) < datetime.utcnow()

    def get_id(self):
        return self.get_pem_encoding()

    def get_pem_encoding(self):
        return self._pem_certificate

    @classmethod
    def from_request(cls, request):
        # HIPVM-251 workaround
        if request.getpeercert(True) is None:
            logger.info("CERT IS NONE")
            address = request.getpeername()[0]
            cert = HIPVM_251_WORKAROUND.get(address)
            # TODO notAfter should be extracted from cert
            return cls(cert, {'notAfter': 'Feb 16 16:54:50 2025 GMT'})
        import ssl
        logger.info("CERT IS NOT NONE")
        der_cert = request.getpeercert(True)
        return cls(ssl.DER_cert_to_PEM_cert(der_cert) if der_cert is not None else None, request.getpeercert(False))

    @classmethod
    def base64_encoded_der_to_pem_format(cls, der_cert):
        from itertools_recipes import grouper
        der_cert = ''.join(der_cert.splitlines())
        lines = ['-----BEGIN CERTIFICATE-----']
        lines.extend([''.join(item) for item in grouper(64, der_cert, '')])
        lines.append('-----END CERTIFICATE-----')
        lines.append('')
        pem_cert = '\n'.join(lines)
        return pem_cert

    @classmethod
    def from_vasa_api(cls, newCertificate):
        # The certificate string is not required to be in URL-safe mode.
        pem_cert = cls.base64_encoded_der_to_pem_format(newCertificate)
        return cls(pem_cert, None)


class TrustedStore(object):
    _trusted_ids = set()
    _last_revoked_certificate_id = None

    @classmethod
    def is_trusted(cls, certificate):
        return certificate.get_id() is not None and certificate.get_id() in cls._trusted_ids

    @classmethod
    def is_revoked(cls, certificate):
        return certificate.get_id() is not None and certificate.get_id() == cls._last_revoked_certificate_id

    @classmethod
    def add(cls, certificate):
        if certificate.get_id() is None:
            return
        cls._last_revoked_certificate_id = None
        cls._trusted_ids.add(certificate.get_id())
        cls._append_certificate_to_ca_certs(certificate)
        cls._hipvm_251_workaround__add(certificate)

    @classmethod
    def remove(cls, certificate):
        if not cls.is_trusted(certificate):
            return
        cls._trusted_ids.discard(certificate.get_id())
        cls._last_revoked_certificate_id = certificate.get_id()
        cls._hipvm_251_workaround__remove()

    @classmethod
    def _hipvm_251_workaround__add(cls, certificate):
        from .local_storage import get_request
        key = get_request().getpeername()[0]
        HIPVM_251_WORKAROUND.add(key, certificate.get_id())

    @classmethod
    def _hipvm_251_workaround__remove(cls):
        from .local_storage import get_request
        key = get_request().getpeername()[0]
        HIPVM_251_WORKAROUND.remove(key)

    @classmethod
    def _append_certificate_to_ca_certs(cls, certificate):
        import os
        from shutil import move, copy
        from tempfile import mkstemp
        from ...config import Configuration
        fd, ca_certs_temp = mkstemp()
        os.close(fd)
        ca_certs = Configuration.get_webserver_ssl_configuration()['ca_certs']
        copy(ca_certs, ca_certs_temp)
        with open(ca_certs_temp, 'a') as fd:
            fd.write(certificate.get_pem_encoding())
        move(ca_certs_temp, ca_certs)
