from logging import getLogger
logger = getLogger(__name__)

class Session(object):
    sessions_data = {}

    def __init__(self, certificate_id, session_id):
        super(Session, self).__init__()
        self._certificate_id = certificate_id
        self._session_id = session_id
        if self.get_timeout_in_seconds() is None:
            self.set_session_timeout(ActiveSessionsPool.TIMEOUT_IN_SECONDS)

    def get_id(self):
        if self._certificate_id is None or self._session_id in (None, 0):
            return None
        return (None, self._session_id)

    def get_session_id(self):
        return self._session_id

    def set_session_timeout(self, seconds):
        self.set_session_value('_session_timeout', seconds)

    def get_timeout_in_seconds(self):
        return self.get_session_value('_session_timeout')

    def get_session_value(self, key):
        return self.sessions_data.get(self._session_id, dict()).get(key, None)

    def set_session_value(self, key, value):
        self.sessions_data.setdefault(self._session_id, dict())[key] = value

    @classmethod
    def from_handler(cls, handler):
        from . import client_certificates
        certificate_id = client_certificates.Certificate.from_request(handler.request).get_id()
        cookies = handler.headers.get("cookie", 'VASASESSIONID=0')
        [session_id] = [cookie.split('=')[1].strip() for cookie in cookies.split(';')
                        if cookie.split('=')[0].strip() == "VASASESSIONID"]
        return cls(certificate_id, int(session_id))

    @classmethod
    def generate_new(cls):
        from . import local_storage
        from . import client_certificates
        request = local_storage.get_request()
        certificate_id = client_certificates.Certificate.from_request(request).get_id()
        session_id = ActiveSessionsPool._session_counter.next()
        return cls(certificate_id, session_id)

import itertools

class ActiveSessionsPool(object):
    TIMEOUT_IN_SECONDS = 60 * 10
    _trusted_ids = dict()
    _session_counter = itertools.count(start=1)

    @classmethod
    def is_trusted(cls, session):
        return session.get_id() in cls._trusted_ids

    @classmethod
    def is_expired(cls, session):
        from datetime import datetime
        return cls._trusted_ids[session.get_id()] < datetime.now()

    @classmethod
    def add(cls, session):
        from datetime import datetime, timedelta
        if session.get_id() is not None:
            cls._trusted_ids[session.get_id()] = datetime.now() + timedelta(seconds=session.get_timeout_in_seconds())

    @classmethod
    def update(cls, session):
        cls.add(session)

    @classmethod
    def remove(cls, session):
        if session and session.get_id() in cls._trusted_ids:
            cls._trusted_ids.pop(session.get_id())


def get_session():
    from . import local_storage
    session = local_storage.get_session()
    if session is None:
        handler = local_storage.get_handler()
        session = Session.from_handler(handler)
        local_storage.set_session(session)
    return session
