from infi.storagemodel.vendor.infinidat.infinibox import vid_pid, vid_pid_with_spaces
from infi.pyvmomi_wrapper.esxcli import EsxCLI
from infi.pyutils.decorators import wraps
from munch import Munch
from logging import getLogger

logger = getLogger(__name__)

def function_logger(func):
    @wraps(func)
    def callee(*args, **kwargs):
        logger.info("Calling {!r}".format(func.__name__))
        try:
            return_value = func(*args, **kwargs)
            logger.info("Call to {!r} returned: {!r}".format(func.__name__, return_value))
            return return_value
        except:
            logger.exception("{!r} raised an exception".format(__name__))
            raise
    return callee

class InfinidatSatp(object):
    vid_pid = vid_pid
    vid_pid_with_spaces = vid_pid_with_spaces
    INFINIDAT_RULE_PROPERTIES = Munch(satp="VMW_SATP_ALUA", psp="VMW_PSP_RR", claimoption="tpgs_on", pspoption="iops=1",
                                      vendor=vid_pid[0], model=vid_pid[1])

    def __init__(self, esx_host):
        self._esx_host = esx_host
        self._esxcli = EsxCLI(self._esx_host)

    @classmethod
    def _get_rule_properties(cls, rule):
        from munch import Munch
        RULE_MAPPINGS = {'Name': 'satp', 'DefaultPSP': 'psp', 'ClaimOptions': 'claimoption', 'PSPOptions': 'pspoption',
                         'Vendor': 'vendor', "Model": 'model', 'Driver': 'driver', 'Description': 'description'}
        properties = {value: getattr(rule, key) for key, value in RULE_MAPPINGS.items()
                      if getattr(rule, key) is not None}
        return Munch(properties)

    def _get_nmp_satp_rule_handler(self):
        return self._esxcli.get("storage.nmp.satp.rule")

    def _get_nmp_device_handler(self):
        return self._esxcli.get('storage.nmp.device')

    def _get_nmp_roundrobin_handler(self):
        return self._esxcli.get('storage.nmp.psp.roundrobin.deviceconfig')

    def _get_core_device_handler(self):
        return self._esxcli.get('storage.core.device')

    def add_rule(self):
        rule_handler = self._get_nmp_satp_rule_handler()
        kwargs = self.INFINIDAT_RULE_PROPERTIES.copy()
        kwargs.update(dict(boot=True, description="INFINIDAT"))
        rule_handler.Add(**kwargs)

    @function_logger
    def get_rules(self):
        rule_handler = self._get_nmp_satp_rule_handler()
        return [self._get_rule_properties(rule) for rule in rule_handler.List()]

    def _remove_existing_rule(self, rule):
        kwargs = rule.copy()
        kwargs.update(dict(boot=True))
        logger.info("Deleting rule {!r}".format(kwargs))
        rule_handler = self._get_nmp_satp_rule_handler()
        try:
            rule_handler.Remove(**kwargs)
        except:
            logger.exception("Failed to delete rule")
            raise

    def _is_infinidat_rule(self, rule):
        return (rule.get('vendor'), rule.get('model')) == self.vid_pid

    def remove_rules(self):
        for rule in [rule for rule in self.get_rules() if self._is_infinidat_rule(rule)]:
            self._remove_existing_rule(rule)

    @function_logger
    def is_rule_installed(self):
        return any(all(value == rule.get(key) for key, value in self.INFINIDAT_RULE_PROPERTIES.items())
                   for rule in self.get_rules())

    @function_logger
    def _get_infinidat_devices_from_core_handler(self):
        handler = self._get_core_device_handler()
        return [device for device in handler.List() if (device.Vendor, device.Model) == self.vid_pid_with_spaces]

    @function_logger
    def _get_all_devices_from_nmp_handler(self):
        handler = self._get_nmp_device_handler()
        return [device for device in handler.List()]

    @function_logger
    def _refresh_policies_for_existing_infinidat_devices__generator(self):
        round_robin_handler = self._get_nmp_roundrobin_handler()
        nmp_handler = self._get_nmp_device_handler()
        for device in self._get_infinidat_devices_from_core_handler():
            nmp_handler.Set(device=device.Device, psp=self.INFINIDAT_RULE_PROPERTIES.psp)
            round_robin_handler.Set(device=device.Device, iops=1, type="iops")
            yield

    @function_logger
    def fix_everything(self):
        from infi.vendata.vmware_powertools.vcenter.shortcuts import create_vcenter_task
        with create_vcenter_task(self._esx_host, "Host Readiness") as task:
            percentage = 0
            increment = 100 / (2+len(self._get_infinidat_devices_from_core_handler()))
            self.remove_rules()
            percentage += increment
            task.update_progress(percentage)
            self.add_rule()
            percentage += increment
            task.update_progress(percentage)
            for run in self._refresh_policies_for_existing_infinidat_devices__generator():
                percentage += increment
                task.update_progress(percentage)

    def _get_matching_nmp_device_settings(self, device):
        for item in self._get_all_devices_from_nmp_handler():
            if item.Device == device.Device:
                return item

    def get_nmp_settings_for_infinibox_devices(self):
        core_devices = {item.Device: item for item in self._get_infinidat_devices_from_core_handler()}
        return {item.Device: item for item in self._get_all_devices_from_nmp_handler()
               if item.Device in core_devices}

    @function_logger
    def are_all_devices_with_correct_settings(self):
        round_robin_handler = self._get_nmp_roundrobin_handler()
        devices = self.get_nmp_settings_for_infinibox_devices()
        expected = (self.INFINIDAT_RULE_PROPERTIES.satp, self.INFINIDAT_RULE_PROPERTIES.psp)
        if not all((value.StorageArrayType, value.PathSelectionPolicy) == expected for value in devices.values()):
            return False
        for key in devices:
            round_robin_settings = round_robin_handler.Get(device=key)
            if round_robin_settings.IOOperationLimit != 1:
                return False
            if round_robin_settings.LimitType.lower() not in ('iops', 'default'):
                return False
        return True
