from infi.pyutils.lazy import cached_method
from mock import MagicMock
from logging import getLogger
from pyVmomi import vim

from .managed_object_factory import ManagedObjectFactory
from .patches import AppliancePatches, StorageModelPatches, PyvisdkPatches, IntegrationTestsPatches
from ..patches import ContextQueue

logger = getLogger(__name__)

class vClient(object):  # pragma: no cover
    def __init__(self, vmock):
        super(vClient, self).__init__()
        self.client = MagicMock()
        self.client.factory.create = self._create_data_object
        self._vmock = vmock
        self.service_instance = self._vmock.service_instance
        self.service_content = self.service_instance.content
        self.session_manager = MagicMock()
        self.session_manager.currentSession.key = '67c8af4c-a2ec-11e3-bc7a-84383565d680'
        self.root = self.service_instance.content.rootFolder
        self._connected = False
        self._logged_in = False
        self.property_collectors = MagicMock()

    def logout(self):
        pass

    def _create_data_object(self, namespace):
        data_object = MagicMock()
        data_object.__repr__ = lambda zelf: "<vMock DataObject: {}>".format(__name__)
        return data_object

    def connect(self):
        self._connected = True

    def login_extension_by_certificate(self, *args, **kwargs):
        self._logged_in = True

    def get_reference_to_managed_object(self, managed_object):
        return managed_object._moref

    def get_managed_object_by_reference(self, moref):
        return self._vmock.get_managed_object_by_reference(moref)

    def get_host_systems(self):
        return self.get_decendents_by_name(vim.HostSystem)

    def get_host_system(self, esx_host_name):
        return [managed_object for managed_object in self.get_host_systems()
                if managed_object.name == esx_host_name][0]

    def get_virtual_machines(self):
        return self.get_decendents_by_name(vim.VirtualMachine)

    def get_folders(self):
        return self.get_decendents_by_name(vim.Folder)

    def get_datacenters(self):
        return self.get_decendents_by_name(vim.Datacenter)

    def get_host_clusters(self):
        return self.get_decendents_by_name(vim.ClusterComputeResource)

    def get_datastores(self):
        datastores = list()
        for datacenter in self.get_datacenters():
            datastores.extend(datacenter.datastore)
        return datastores

    def get_decendents_by_name(self, _type, *args, **kwargs):
        managed_objects = self._vmock.get_managed_objects()
        return [value for key, value in managed_objects.items() if key.startswith(_type.__name__.split(".")[-1])]

    def wait_for_task(self, task, timeout=None):
        return "success"

    def get_virtual_machine(self, name):
        matching = [vm for vm in self.get_virtual_machines() if vm.name == name]
        if len(matching) == 0:
            return None
        return matching[0]

class vMock(object):  # pragma: no cover
    factory = ManagedObjectFactory

    def __init__(self):
        super(vMock, self).__init__()
        self._moref_id_counter = 1
        self._managed_objects = {}
        self._vm_uuids = {}
        self._queue = ContextQueue()
        self.factory.create_service_instance(self)
        self.factory.create_singletons(self)
        self._simulators = []
        self._infiniboxes = []
        self._device_names = {}

    def __enter__(self):
        self._apply_monkey_patching()

    def _apply_monkey_patching(self):
        self.storagemodel_patchers = StorageModelPatches(self)
        self.storagemodel_patchers.start()
        self.appliance_patchers = AppliancePatches(self)
        self.appliance_patchers.start()
        self.pyvisdk_patchers = PyvisdkPatches(self)
        self.pyvisdk_patchers.start()
        self.integration_tests_patchers = IntegrationTestsPatches(self)
        self.integration_tests_patchers.start()

    def __exit__(self, *exc_info):
        exception_has_been_raised = exc_info != (None, None, None)
        self._stop_monkey_patching()
        if exception_has_been_raised:
            logger.error("Exception has been raised", exc_info=exc_info)
            raise exc_info[0], exc_info[1], exc_info[2]

    def _stop_monkey_patching(self):
        self.appliance_patchers.stop()
        self.storagemodel_patchers.stop()
        self.pyvisdk_patchers.stop()
        self.integration_tests_patchers.stop()

    def get_managed_object_by_reference(self, moref):
        try:
            return self._managed_objects[moref]
        except KeyError, e:
            raise ValueError(e.message)

    @cached_method
    def get_client(self):
        client = vClient(self)
        self._client = client
        return client

    def get_managed_objects(self):
        return dict(self._managed_objects)

class vMockFactory(object):  # pragma: no cover
    @classmethod
    def _split_hosts_into_cluster_groups(cls, hosts, clusters):
        return [hosts[i::clusters] for i in xrange(clusters)]

    @classmethod
    def _add_initiator_port_to_host(cls, host):
        host._smock.get_inventory().add_initiator()

    @classmethod
    def create_infinibox(cls, vmock):
        from infinisim.infinibox import Infinibox as Simulator
        from infinisdk import InfiniBox
        simulator = Simulator()
        simulator.activate()
        infinibox = InfiniBox(simulator, auth=('admin', '123456'))
        infinibox.pools.create()
        vmock._simulators.append(simulator)
        vmock._infiniboxes.append(infinibox)
        return infinibox, simulator

    @classmethod
    def zone_with_system__best_practices(cls, host, infinibox):
        host._smock.get_inventory().register_infinibox(infinibox)
        host._smock.get_inventory().zone_with_system__best_practices(infinibox)

    @classmethod
    def zone_host_with_infinibox__full_mesh(cls, host, infinibox):
        host._smock.get_inventory().register_infinibox(infinibox)
        host._smock.get_inventory().zone_with_system__full_mesh(infinibox)

    @classmethod
    def clear_infinibox_and_host_relationship(cls, host, infinibox):
        item_to_remove = [system for system in host._smock.get_inventory()._systems
                          if system.get_serial() == infinibox.get_serial()]
        if item_to_remove:
            host._smock.get_inventory()._systems.remove(item_to_remove[0])
        for initiator in host._smock.get_inventory().get_initiator_addresses():
            for target in infinibox.components.fc_ports.get_online_target_addresses():
                for zone in host._smock.get_inventory()._zones:
                    if zone.initiator == initiator and zone.target == target:
                        host._smock.get_inventory()._zones.remove(zone)

    @classmethod
    def purge_all_related_zones_on_host(cls, host):
        for initiator in host._smock.get_inventory().get_initiator_addresses():
            for zone in host._smock.get_inventory()._zones:
                if zone.initiator == initiator:
                    host._smock.get_inventory()._zones.remove(zone)

    @classmethod
    def purge_all_related_zones_from_active_host(cls, vmock):
        from infi.storagemodel.vmware.patches.storagemodel import StorageModelFactory
        cls.purge_all_related_zones_on_host(StorageModelFactory.get()._host)

    @classmethod
    def create(cls, hosts=6, clusters=3, infiniboxes=3, initiator_ports=2, datastores=3, virtual_machines=9,
               map_to_lun0=False):
        from itertools import count
        vmock = vMock()
        with vmock:
            from infi.vendata.vmware_powertools.config import Configuration
            from infi.vendata.vmware_powertools.redis import RedisServer
            RedisServer.start()
            Configuration.populate_defaults(production=False)
        datacenter = vmock.factory.create_datacenter(vmock)
        hosts = [vmock.factory.create_host(vmock, datacenter) for i in range(hosts)]
        [[cls._add_initiator_port_to_host(host) for i in range(initiator_ports)]
         for host in hosts]
        hosts_per_cluster = cls._split_hosts_into_cluster_groups(hosts, clusters)
        clusters = [vmock.factory.create_cluster(vmock, datacenter, hosts_per_cluster[index])
                    for index in range(clusters)]
        infiniboxes_and_simulators = [cls.create_infinibox(vmock) for i in range(infiniboxes)]
        infiniboxes = [_tuple[0] for _tuple in infiniboxes_and_simulators]
        [[cls.zone_with_system__best_practices(host, infinibox)
          for host in hosts] for infinibox in infiniboxes]

        folder = vmock.factory.create_folder(vmock, datacenter.vmFolder)
        if hosts and datastores:
            from infi.vendata.vmware_powertools.utils import auth
            from infi.vendata.credentials_store.base import Credentials
            auth.auth.CREDENTIALS_STORE.set_credentials_for_extension_client(Credentials('admin', '123456'))
            lun = count(0 if map_to_lun0 else 1)
            for i in range(datastores):
                vms = []
                for j in range(virtual_machines/datastores):
                    vms.append(vmock.factory.create_virtual_machine(vmock, hosts[i % len(hosts)], folder))
                with vmock:
                    vmock.factory.create_datastore(vmock, [hosts[i % len(hosts)]], vms,
                                                   infiniboxes[i % len(infiniboxes)], lun.next())
        return vmock
