from pyVmomi import vim
from mock import MagicMock
from munch import Munch
from logging import getLogger
from uuid import uuid1
from infi.vendata.vmware_powertools.vcenter.shortcuts import object_to_moref

logger = getLogger(__name__)

MANAGED_OBJECT_SUBTYPES = {
    'HostSystem': 'host',
    'VirtualMachine': 'vm',
    'Datastore': 'datastore',
    'ClusterComputeResource': 'domain',
    'ComputeResource': 'domain',
    'HostStorageSystem': 'storageSystem',
    'HostDatastoreSystem': 'datastoreSystem',
    'Datacenter': 'datacenter',
    'Folder': 'group',
    'Task': 'task'
}

class ManagedObjectFactory(object):  # pragma: no cover
    @classmethod
    def create_singleton_object(cls, vmock, name):
        singleton = MagicMock()
        singleton.__repr__ = lambda zelf: "<vMock Singleton: {}>".format(name)
        singleton._vmock = vmock
        return singleton

    @classmethod
    def get_and_increase_id(cls, vmock, moref_id_counter=None):
        with vmock._queue.wait_for_your_turn():
            _id = moref_id_counter or vmock._moref_id_counter
            if moref_id_counter is None:
                vmock._moref_id_counter += 1
            return _id

    @classmethod
    def create_managed_object(cls, vmock, managed_object_type, moref_id_counter=None):
        moid = "{}-{}".format(MANAGED_OBJECT_SUBTYPES[managed_object_type],
                              cls.get_and_increase_id(vmock, moref_id_counter))
        moref = "{}:{}".format(managed_object_type, moid)
        managed_object = MagicMock(spec=getattr(vim, managed_object_type))
        managed_object._moId = moid
        managed_object.name = moid
        managed_object._moref_id_counter = moref_id_counter
        managed_object._moref = moref
        managed_object.availableField = []
        managed_object.customValue = []
        managed_object.__repr__ = lambda zelf: "<vMock ManagedObject: {}>".format(moref)
        managed_object._vmock = vmock
        def Destroy_Task(*args, **kargs):
            vmock._managed_objects.pop(moref)
            return cls.create_task(vmock)
        managed_object.Destroy_Task = Destroy_Task
        vmock._managed_objects[moref] = managed_object
        return managed_object

    @classmethod
    def create_host(cls, vmock, parent):
        """Creates an ESX host
        :param parent: datacenter or cluster
        """
        from infi.vendata.smock import HostMock
        host = cls.create_managed_object(vmock, "HostSystem")
        host.runtime.connectionState = "connected"
        host.runtime.powerState = "poweredOn"
        host.runtime.inMaintenanceMode = False
        hostname = host.name
        host.summary.hardware.uuid = str(uuid1())
        host.summary.config.name = hostname
        host.summary.config.product.name = "product.name"
        host.summary.config.product.fullName = "product.fullName"
        host.datastore = []
        host.parent = parent
        host.vm = []

        smock = HostMock(False)
        host._smock = smock
        smock._clear_storagemodel_cache = lambda: True
        inventory = smock.get_inventory()
        inventory._hostname = hostname

        storage_system = cls.create_managed_object(vmock, "HostStorageSystem", host._moref_id_counter)
        datastore_system = cls.create_managed_object(vmock, "HostDatastoreSystem", host._moref_id_counter)

        host.configManager.storageSystem = storage_system
        host.configManager.datastoreSystem = datastore_system

        host._readiness_state = False

        def UpdateScsiLunDisplayName(*args, **kwargs):
            lun_uuid, display_name = args
            vmock._device_names[lun_uuid] = display_name

        def QueryVmfsDatastoreCreateOptions(*args, **kwargs):
            device_path, major_version = args
            create_options = MagicMock()
            from infi.storagemodel import get_storage_model
            device = [device
                      for device in get_storage_model().get_native_multipath().get_all_multipath_block_devices()
                      if device.get_block_access_path() == device_path][0]
            create_options.spec._disk_name = device.get_canonical_name()
            return [create_options]

        def CreateVmfsDatastore(*args, **kwargs):
            create_options_spec, = args
            datastore = cls.create_datastore(vmock)
            datastore.host.append(Munch(key=host,
                                        mountInfo=Munch(accessible=True, mounted=True,
                                                        path="/vmfs/volumes/{}".format(datastore.info.vmfs.uuid))))
            datastore.info.vmfs.extent.append(Munch(diskName=create_options_spec._disk_name,
                                                    partition=1))
            datastore.info.name = create_options_spec.vmfs.volumeName
            datastore.summary.name = datastore.info.name
            host.datastore.append(datastore)
            return datastore

        def ResignatureUnresolvedVmfsVolume_Task(*args, **kwargs):
            from infi.vendata.vmware_powertools.patches import with_host
            spec, = args
            device_path = spec.extentDevicePath.rsplit(':', 1)[0]
            with with_host(host):
                create_options_spec = QueryVmfsDatastoreCreateOptions(device_path, 5)[0].spec
            datastore = CreateVmfsDatastore(create_options_spec)
            task = cls.create_task(vmock)
            task.info.result.result = datastore
            return task

        def UnmountVmfsVolume(*args, **kwargs):
            vmfs_uuid, = args
            datastore = [datastore
                         for datastore in host.datastore
                         if datastore.info.vmfs.uuid == vmfs_uuid][0]
            mount_info = [item
                          for item in datastore.host
                          if item.key._moId == host._moId][0]
            mount_info.mountInfo.mounted = False
            host.datastore.remove(datastore)
            datastore.host.remove(mount_info)

        def MountVmfsVolume(*args, **kwargs):
            vmfs_uuid, = args
            datastore = [datastore
                         for datastore in host.datastore
                         if datastore.info.vmfs.uuid == vmfs_uuid][0]
            datastore.host.append(Munch(key=host,
                                        mountInfo=Munch(accessible=True, mounted=True,
                                                        path="/vmfs/volumes/{}".format(datastore.info.vmfs.uuid))))

        def QueryVmfsDatastoreExpandOptions(*args, **kwargs):
            from pyVmomi import vim
            info = vim.VmfsDatastoreSingleExtentOption()
            spec = vim.VmfsDatastoreExpandSpec()
            return [Munch(info=info, spec=spec)]

        def ExpandVmfsDatastore(*args, **kwargs):
            pass

        def DetachScsiLun(lun_uuid):
            from infi.vendata.vmware_powertools.patches import with_host
            with with_host(host):
                from infi.storagemodel import get_storage_model
                devices = get_storage_model().get_native_multipath().get_all_multipath_block_devices()
                [devices.remove(item) for item in devices[:] if item.get_uuid() == lun_uuid]

        datastore_system.QueryVmfsDatastoreCreateOptions = QueryVmfsDatastoreCreateOptions
        datastore_system.CreateVmfsDatastore = CreateVmfsDatastore
        datastore_system.ResignatureUnresolvedVmfsVolume_Task = ResignatureUnresolvedVmfsVolume_Task
        datastore_system.QueryVmfsDatastoreExpandOptions = QueryVmfsDatastoreExpandOptions
        datastore_system.ExpandVmfsDatastore = ExpandVmfsDatastore
        storage_system.UnmountVmfsVolume = UnmountVmfsVolume
        storage_system.UpdateScsiLunDisplayName = UpdateScsiLunDisplayName
        storage_system.MountVmfsVolume = MountVmfsVolume
        storage_system.DetachScsiLun = DetachScsiLun
        if parent._moref.startswith("Datacenter"):
            domain = cls.create_compute_resource(vmock, parent, host)
            parent.hostFolder.childEntity.append(domain)
            host.parent = domain
        elif parent._moref.startswith("ClusterComputeResource"):
            parent.host.append(host)
            host.parent = parent
        else:
            raise AttributeError("parent must be either a datacenter or a cluster")
        return host

    @classmethod
    def create_cluster(cls, vmock, datacenter, hosts=[]):
        cluster = cls.create_managed_object(vmock, "ClusterComputeResource")
        datacenter.hostFolder.childEntity.append(cluster)
        cluster.parent = datacenter.hostFolder
        cluster.host = list([host for host in hosts if host is not None])
        for host in hosts:
            if host.parent._moref.startswith("ComputeResource"):
                # Remove stale ComputeResource from Datacenter
                datacenter = host.parent.parent
                datacenter.hostFolder.childEntity.remove(host.parent)
            elif host.parent._moref.startswith("ClusterComputeResource"):
                raise ValueError("host {!r} is already a member in another cluster".format(host))
            host.parent = cluster
        class Datastores(object):
            def __iter__(self):
                all_datastores = []
                for host in cluster.host:
                    all_datastores += host.datastore
                return list(set(all_datastores)).__iter__()
            def __len__(self):
                return len([item for item in self])
        cluster.datastore = Datastores()
        return cluster

    @classmethod
    def create_compute_resource(cls, vmock, datacenter, host):
        domain = cls.create_managed_object(vmock, "ComputeResource")
        domain.parent = datacenter
        domain.host = [host]
        class Datastores(object):
            def __iter__(self):
                all_datastores = []
                for host in domain.host:
                    all_datastores += host.datastore
                return list(set(all_datastores)).__iter__()
            def __len__(self):
                return len([item for item in self])
        domain.datastore = Datastores()
        return domain

    @classmethod
    def create_folder(cls, vmock, parent):
        moref_id = None if parent else 'd1'
        folder = cls.create_managed_object(vmock, "Folder", moref_id)
        folder.childEntity = []
        folder.parent = parent if parent else None
        def RegisterVM_Task(*args, **kwargs):
            import re
            datastore_path, = args
            name = kwargs["name"]
            host = kwargs["host"]
            vm = cls.create_virtual_machine(vmock, host, folder)
            original_vm_name = re.search("\[.*?]\ (.*?)/.*?", datastore_path).group(1)
            try:
                vm.config.uuid = vm.summary.config.uuid = vmock._vm_uuids[original_vm_name]
            except KeyError:
                pass
            task = cls.create_task(vmock)
            vm.config.files.vmPathName = datastore_path.replace(original_vm_name, name)
            task.info.result = vm
            vm.name = name
            return task
        folder.RegisterVM_Task = RegisterVM_Task
        return folder

    @classmethod
    def create_datacenter(cls, vmock):
        datacenters_folder = vmock.get_managed_object_by_reference("Folder:group-d1")
        datacenter = cls.create_managed_object(vmock, "Datacenter")
        datacenter.parent = datacenters_folder
        datacenter.datastoreFolder = cls.create_folder(vmock, datacenter)
        datacenter.hostFolder = cls.create_folder(vmock, datacenter)
        datacenter.vmFolder = cls.create_folder(vmock, datacenter)
        datacenters_folder.childEntity.append(datacenter)
        class Datastores(object):
            def __iter__(self):
                all_datastores = []
                for computer_resource in datacenter.hostFolder.childEntity:
                    all_datastores += computer_resource.datastore
                return list(set(all_datastores)).__iter__()
            def __len__(self):
                return len([item for item in self])
        datacenter.datastore = Datastores()
        return datacenter

    @classmethod
    def create_service_instance(cls, vmock):
        service_instance = cls.create_singleton_object(vmock, "ServiceInstance")
        service_instance.content = MagicMock()
        service_instance.content.about.instanceUuid = str(uuid1())
        vmock.service_instance = service_instance

    @classmethod
    def create_singletons(cls, vmock):
        cls.create_root_folder(vmock)
        cls.create_extension_manager(vmock)
        cls.create_custom_fields_manager(vmock)
        cls.create_task_manager(vmock)

    @classmethod
    def create_root_folder(cls, vmock):
        root_folder = cls.create_folder(vmock, None)
        root_folder.name = "Datacenters"
        vmock.service_instance.content.rootFolder = root_folder

    @classmethod
    def create_extension_manager(cls, vmock):
        extension_manager = cls.create_singleton_object(vmock, "ExtensionManager")
        extension_manager.extensionList = []
        def get_extensions_by_key():
            return {extension.key: extension
                                   for extension in extension_manager.extensionList}
        def UpdateExtension(*args, **kwargs):
            with vmock._queue.wait_for_your_turn():
                extension, = args
                extensions = get_extensions_by_key()
                extensions[extension.key] = extension
                extension_manager.extensionList = extensions.values()
        def RegisterExtension(*args, **kwargs):
            UpdateExtension(*args, **kwargs)
        def SetExtensionCertificate(*args, **kwargs):
            pass
        def FindExtension(*args, **kwargs):
            key, = args
            return get_extensions_by_key()[key]
        extension_manager.UpdateExtension = UpdateExtension
        extension_manager.RegisterExtension = RegisterExtension
        extension_manager.SetExtensionCertificate = SetExtensionCertificate
        extension_manager.FindExtension = FindExtension
        vmock.service_instance.content.extensionManager = extension_manager

    @classmethod
    def create_custom_fields_manager(cls, vmock):
        custom_fields_manager = cls.create_singleton_object(vmock, "CustomFieldsManager")
        custom_fields_manager._counter = 101
        custom_fields_manager.field = []
        def AddCustomFieldDef(name, moType, fieldDefPolicy=None, fieldPolicy=None):
            field = Munch(key=custom_fields_manager._counter, type="string",
                          name=name, managedObjectType=moType)
            custom_fields_manager.field.append(field)
            custom_fields_manager._counter += 1
        def SetField(entity, key, value):
            with vmock._queue.wait_for_your_turn():
                custom_values = {item.key: item for item in entity.customValue}
                custom_values[key] = Munch(key=key, value=value)
                entity.customValue = custom_values.values()
        custom_fields_manager.AddCustomFieldDef = AddCustomFieldDef
        custom_fields_manager.SetField = SetField
        vmock.service_instance.content.customFieldsManager = custom_fields_manager

    @classmethod
    def create_task_manager(cls, vmock):
        task_manager = cls.create_singleton_object(vmock, "TaskManager")
        def CreateTask(*args, **kwargs):
            task = cls.create_task(vmock)
            return task
        task_manager.CreateTask = CreateTask
        vmock.service_instance.content.taskManager = task_manager

    @classmethod
    def create_task(cls, vmock):
        task = cls.create_managed_object(vmock, "Task")
        task.info.key = ''
        task.task = task
        return task

    @classmethod
    def create_virtual_machine(cls, vmock, host, folder):
        import uuid
        vm = cls.create_managed_object(vmock, "VirtualMachine")
        vm.datastore = []
        vm.config.hardware.device = []
        vm.parent = folder
        vm.config.uuid = str(uuid.uuid1())
        vm.summary.config.uuid = vm.config.uuid
        vmock._vm_uuids[vm.name] = vm.config.uuid
        vm.summary.runtime.host = host
        folder.childEntity.append(vm)
        host.vm.append(vm)
        def RelocateVM_Task(*args, **kwargs):
            relocate_spec, _ = args
            new_datastore_ref = relocate_spec.datastore
            new_datastore_ref = object_to_moref(new_datastore_ref)
            new_datastore = vmock.get_client().get_managed_object_by_reference(new_datastore_ref)
            for old_datastore in vm.datastore:
                old_datastore.vm.remove(vm)
            new_datastore.vm.append(vm)
            vm.datastore = [new_datastore]
            vm.config.hardware.device = vim.VirtualDisk(backing=vim.VirtualDiskFlatVer2BackingInfo(datastore=new_datastore))
            task = cls.create_task(vmock)
            return task
        def Destroy_Task(*args, **kargs):
            vmock._managed_objects.pop(object_to_moref(vm))
            [ds.vm.remove(vm) for ds in vm.datastore]
            vm.summary.runtime.host.vm.remove(vm)
            vm.parent.childEntity.remove(vm)
            return cls.create_task(vmock)
        def ReconfigVM_Task(*args, **kwargs):
            cfg_spec, = args
            vm.summary.config.uuid = vm.config.uuid = cfg_spec.uuid
            return cls.create_task(vmock)
        vm.Destroy_Task = Destroy_Task
        vm.RelocateVM_Task = RelocateVM_Task
        vm.ReconfigVM_Task = ReconfigVM_Task
        return vm

    @classmethod
    def _register_host(cls, vmock, infinibox, host):
        from infi.vendata.vmware_powertools.webapp.plugin.views.host.registration import register_host
        from infi.vendata.vmware_powertools.patches import with_host
        register_host(host, infinibox.get_serial())
        with with_host(host):
            from infi.storagemodel import get_storage_model
            get_storage_model().rescan_and_wait_for()
        # we get the host by name because getting the id takes too long and the registration registers the name ok
        system_host = infinibox.hosts.get(name=host.name)
        return system_host

    @classmethod
    def _map_volume_to_host(cls, vmock, infinibox, volume, host, lun=None):
        from infi.vendata.vmware_powertools.snapshots.revert import wait_for_volume, find_volume_device
        system_host = cls._register_host(vmock, infinibox, host)
        system_host.map_volume(volume, lun=lun)
        wait_for_volume(host, volume)
        return find_volume_device(host, volume).get_canonical_name()

    @classmethod
    def add_datastore_to_host(cls, vmock, datastore, host, lun=None):
        disk_name = cls._map_volume_to_host(vmock, datastore._volume.get_system(), datastore._volume, host, lun=lun)
        datastore.info.vmfs.extent.append(Munch(diskName=disk_name,
                                                partition=1))
        host.datastore.append(datastore)

    @classmethod
    def _add_vm_to_datastore(cls, vm, datastore):
        vm.datastore.append(datastore)
        vm.config.hardware.device.append(vim.VirtualDisk(backing=vim.VirtualDiskFlatVer2BackingInfo(datastore=datastore)))
        datastore.vm.append(vm)
        vm.config.files.vmPathName = "[{}] {}/{}.vmx".format(datastore.summary.name, vm.name, vm.name)

    @classmethod
    def create_datastore(cls, vmock, hosts=[], virtual_machines=[], infinibox=None, lun=None):
        from capacity import GiB
        import random
        datastore = cls.create_managed_object(vmock, "Datastore")
        datastore.vm = []
        datastore.host = []
        datastore.summary.type = "VMFS"
        datastore.info.url = str(uuid1())
        datastore.info.vmfs.uuid = str(uuid1())
        datastore.info.vmfs.extent = []
        datastore.capability.storageIORMSupported = False
        for host in hosts:
            host.datastore.append(datastore)
            datastore.host.append(Munch(key=host,
                                        mountInfo=Munch(accessible=True, mounted=True)))
        if infinibox is not None:
            name = ''.join([random.choice('ABCDEFGHIJKLMNOPQRSTUVWXYZ') for i in range(6)])
            datastore._volume = infinibox.volumes.create(pool=infinibox.pools.get_all()[0], size=1*GiB, name=name)
            for host in hosts:
                disk_name = cls._map_volume_to_host(vmock, datastore._volume.get_system(), datastore._volume, host, lun=lun)
                datastore.info.vmfs.extent.append(Munch(diskName=disk_name,
                                                        partition=1))
            datastore.summary.name = datastore._volume.get_name()
            vmock.appliance_patchers._remember_datastore(datastore._volume, datastore)
        for vm in virtual_machines:
            cls._add_vm_to_datastore(vm, datastore)
        def Destroy_Task():
            vmock._managed_objects.pop(object_to_moref(datastore))
            for vm in datastore.vm:
                vm.datastore.remove(datastore)
            for host in datastore.host:
                host.key.datastore.remove(datastore)
            return cls.create_task(vmock)
        datastore.Destroy_Task = Destroy_Task
        def Rename_Task(name):
            datastore.summary.name = name
            return cls.create_task(vmock)
        datastore.Rename_Task = Rename_Task
        return datastore

    @classmethod
    def create_extension(cls, vmock):
        from infi.vendata.vmware_powertools.config import Configuration
        extension = MagicMock()
        extension.client = []
        extension.company = "company"
        extension.description.label = "label"
        extension.description.summary = "summary"
        extension.eventList = []
        extension.extendedProductInfo.companyUrl = ''
        extension.extendedProductInfo.managementUrl = ''
        extension.extendedProductInfo.productUrl = ''
        extension.extendedProductInfo.self = None
        extension.faultList = []
        extension.key = Configuration.get_vcenter_extension_key()
        extension.managedEntityInfo = []
        extension.privilegeList = []
        extension.resourceList = []
        extension.server = []
        extension.subjectName = []
        extension.taskList = []
        extension.type = ''
        extension.version = '1.0'
        vmock.service_instance.content.extensionManager.RegisterExtension(extension)
