from mock import patch, MagicMock
from munch import Munch
from logging import getLogger

logger = getLogger(__name__)

class PatchesCollection(object):  # pragma: no cover
    def __init__(self, vmock):
        super(PatchesCollection, self).__init__()
        self._vmock = vmock
        self.patchers = list()

    def stop(self):
        [one_patch.stop() for one_patch in self.patchers]

    def start(self):
        raise NotImplementedError()

class AppliancePatches(PatchesCollection):  # pragma: no cover

    _datastores = {}  # from (system_serial, volume_id) to datastore instance

    def start(self):
        self._ovf()
        self._configuration()
        self._vmware_vcenter_client_request_hook()
        self._is_host_ready()
        self._fix_esx_host_readiness_state()
        self._assert_safe_to_remove()
        self._send_post_request_to_vcenter()
        self._update_vm_version_if_necessary()
        self._create_vmfs_datastore()
        self._mount_vmfs_datastore()
        self._map_existing_volume()
        self._get_host_by_uuid()
        self._get_vm_by_uuid()
        self._storagemodel_patch()
        self._hbaapi_patch()

    def _ovf(self):
        patcher = patch("infi.vendata.vmware_powertools.ovf.OVF")
        OVF = patcher.start()
        OVF.get_attributes.return_value = dict(IP="127.0.0.1")
        self.patchers.append(patcher)

    def _configuration(self):
        from infi.vendata.vmware_powertools.config import Configuration, DEVELOPMENT_CONFIGURATION
        patcher1 = patch.object(Configuration, "_get_source_address")
        _get_source_address = patcher1.start()
        _get_source_address.return_value = "127.0.0.1"
        patcher2 = patch.object(Configuration, "get_vcenter_extension_key")
        get_vcenter_extension_key = patcher2.start()
        get_vcenter_extension_key.return_value = DEVELOPMENT_CONFIGURATION['vcenter_extension_key']
        self.patchers += [patcher1, patcher2]

    def _vmware_vcenter_client_request_hook(self):
        patcher = patch("infi.vendata.vmware_powertools.webapp.plugin.resource.vmware_vcenter_client_request_hook")
        patcher.start()
        self.patchers.append(patcher)

    def _is_host_ready(self):
        from infi.vendata.vmware_powertools.webapp.plugin.views.host import readiness
        patcher = patch.object(readiness, "is_host_ready")
        def side_effect(*args, **kwargs):
            from infi.vendata.vmware_powertools.webapp.plugin.views.host.readiness import ReadinessException
            esx_host,  = args
            if not esx_host._readiness_state:
                raise ReadinessException("not ready")
        is_host_ready = patcher.start()
        is_host_ready.side_effect = side_effect
        self.patchers.append(patcher)

    def _fix_esx_host_readiness_state(self):
        from infi.vendata.vmware_powertools.webapp.plugin.views.host import readiness
        patcher = patch.object(readiness, "fix_esx_host_readiness_state")
        def side_effect(*args, **kwargs):
            esx_host, = args
            esx_host._readiness_state = True
        fix_esx_host_readiness_state = patcher.start()
        fix_esx_host_readiness_state.side_effect = side_effect
        self.patchers.append(patcher)

    def _assert_safe_to_remove(self):
        from infi.vendata.vmware_powertools.vcenter.unmap import VolumeUnmapBase
        patcher = patch.object(VolumeUnmapBase, "assert_safe_to_remove")
        assert_safe_to_remove = patcher.start()
        assert_safe_to_remove.return_value = True
        self.patchers.append(patcher)

        patcher = patch.object(VolumeUnmapBase, "assert_not_being_used_as_rdm")
        assert_not_being_used_as_rdm = patcher.start()
        assert_not_being_used_as_rdm.return_value = True
        self.patchers.append(patcher)

    def _send_post_request_to_vcenter(self):
        from infi.vendata.vmware_powertools.vcenter import vCenter
        patcher = patch.object(vCenter, "_send_post_request_to_vcenter")
        _send_post_request_to_vcenter = patcher.start()
        def side_effect(*args, **kwargs):
            self._vmock.factory.create_extension(self._vmock)
        _send_post_request_to_vcenter.side_effect = side_effect
        self.patchers.append(patcher)

    def _update_vm_version_if_necessary(self):
        from infi.vendata.vmware_powertools.vcenter import vCenter
        patcher = patch.object(vCenter, "update_vm_version_if_necessary")
        patcher.start()
        self.patchers.append(patcher)

    def _create_vmfs_datastore(self):
        from infi.vendata.vmware_powertools.webapp.plugin.views.host.operations import create_vmfs_datastore as func
        def side_effect(*args, **kwargs):
            esx_host, block_device, volume = args
            datastore = func(esx_host, block_device, volume)
            self._remember_datastore(volume, datastore)
            return datastore
        patcher = patch("infi.vendata.vmware_powertools.webapp.plugin.views.host.operations.create_vmfs_datastore")
        create_vmfs_datastore = patcher.start()
        create_vmfs_datastore.side_effect = side_effect
        self.patchers.append(patcher)

    def _mount_vmfs_datastore(self):
        def side_effect(*args, **kwargs):
            esx_host, block_device = args
            for datacenter in self._vmock.get_client().get_datacenters():
                for datastore in datacenter.datastore:
                    for extent in datastore.info.vmfs.extent:
                        if extent.diskName == block_device.get_canonical_name():
                            if datastore not in esx_host.datastore:
                                esx_host.datastore.append(datastore)
                            datastore.host.append((Munch(key=esx_host,
                                                         mountInfo=Munch(accessible=True, mounted=True))))
                            return
            raise ValueError("Datastore not found in any of the datacenters")
        patcher = patch("infi.vendata.vmware_powertools.webapp.plugin.views.host.operations.mount_vmfs_datastore")
        mount_vmfs_datastore = patcher.start()
        mount_vmfs_datastore.side_effect = side_effect
        self.patchers.append(patcher)

    def _remember_datastore(self, volume, datastore):
        key = (volume.get_system().get_serial(), volume.get_id())
        self._datastores[key] = datastore

    def _create_or_get_datastore(self, system_serial, volume_id):
        key = (system_serial, volume_id)
        datastore = self._datastores.get(key) or self._vmock.factory.create_datastore(self._vmock)
        self._datastores[key] = datastore
        return datastore

    def _map_existing_volume(self):
        def add_datastore_to_host(esx_host, new_datastore, system_serial, volume_id):
            from infi.vendata.vmware_powertools.patches import with_host
            esx_host.datastore.append(new_datastore)
            new_dev = None
            with with_host(esx_host):
                from infi.storagemodel.vendor.infinidat.shortcuts import get_infinidat_block_devices
                for dev in get_infinidat_block_devices():
                    if dev.get_vendor().get_volume_id() == int(volume_id) and \
                       dev.get_vendor().get_system_serial() == int(system_serial):
                        new_dev = dev
                        break
            if new_dev:
                disk_name = new_dev.get_canonical_name()
                new_datastore.info.vmfs.extent.append(Munch(diskName=disk_name, partition=1))
                new_datastore.summary.name = new_dev.get_vendor().get_volume_name()

        def host_side_effect(inst, *args, **kwargs):
            # call original function
            host_patcher.temp_original(inst, *args, **kwargs)
            # existing volumes already have datastores on them so we need to add one to the mock
            system_serial, volume_id = inst._system_serial, inst._volume_id
            new_datastore = self._create_or_get_datastore(system_serial, volume_id)
            add_datastore_to_host(inst._esx_host, new_datastore, system_serial, volume_id)

        def cluster_side_effect(inst, *args, **kwargs):
            # call original function
            cluster_patcher.temp_original(inst, *args, **kwargs)
            # existing volumes already have datastores on them so we need to add one to the mock
            system_serial, volume_id = inst._system_serial, inst._volume_id
            datastore = self._datastores.get((system_serial, volume_id))
            if datastore:
                [add_datastore_to_host(esx_host, datastore, system_serial, volume_id)
                 for esx_host in inst._esx_host.parent.host]

        from infi.vendata.vmware_powertools.vcenter.map import MapVolume
        host_patcher = patch.object(MapVolume, "map_to_host", host_side_effect)
        host_patcher.start()
        self.patchers.append(host_patcher)
        cluster_patcher = patch.object(MapVolume, "map_to_cluster", cluster_side_effect)
        cluster_patcher.start()
        self.patchers.append(cluster_patcher)

    def _get_host_by_uuid(self):
        def side_effect(*args, **kwargs):
            client, uuid = args
            matches = [host for host in client.get_host_systems() if host.summary.hardware.uuid == uuid]
            if len(matches) == 0:
                return None
            return matches[0]
        patcher = patch("infi.vendata.vmware_powertools.vcenter.shortcuts.get_host_by_uuid")
        get_host_by_uuid = patcher.start()
        get_host_by_uuid.side_effect = side_effect
        self.patchers.append(patcher)

    def _get_vm_by_uuid(self):
        def side_effect(*args, **kwargs):
            client, uuid = args
            matches = [vm for vm in client.get_virtual_machines() if vm.config.uuid == uuid]
            if len(matches) == 0:
                return None
            return matches[0]
        patcher = patch("infi.vendata.vmware_powertools.vcenter.shortcuts.get_vm_by_uuid")
        get_vm_by_uuid = patcher.start()
        get_vm_by_uuid.side_effect = side_effect
        self.patchers.append(patcher)

    def _hbaapi_patch(self):
        from infi.storagemodel.vmware.patches.hbaapi import HostSystemPortGeneratorFactory
        @classmethod
        def create(*args, **kwargs):
            cls, client, host = args
            from infi.vendata.smock import MockPortGenerator
            return [MockPortGenerator(host._smock.get_inventory())]
        patcher = patch.object(HostSystemPortGeneratorFactory, "create", new=create)
        patcher.start()
        self.patchers.append(patcher)

    def _storagemodel_patch(self):
        from infi.storagemodel.vmware.patches import storagemodel
        from infi.vendata.smock import MockStorageModel
        def create(*args, **kwargs):
            client, moref = args
            host = self._vmock.get_managed_object_by_reference(moref)
            model = MockStorageModel(host._smock.get_inventory())
            model._host = host
            return model
        patcher = patch.object(storagemodel, "VMwareHostStorageModel", new=create)
        patcher.start()
        self.patchers.append(patcher)

class StorageModelPatches(PatchesCollection):  # pragma: no cover
    @staticmethod
    def get_uuid(inst, *args, **kwargs):
        return inst.get_scsi_serial_number()

    @staticmethod
    def get_canonical_name(inst, *args, **kwargs):
        return inst.get_uuid()

    @staticmethod
    def get_all_multipath_storage_controller_devices(inst, *args, **kwargs):
        from infi.vendata.smock.storagemodel.multipath import MockInfiniBoxMultipathStorageController
        from infi.storagemodel import get_storage_model
        multipath_devices = []
        scsi_devices = get_storage_model().get_scsi()._get_all_infinidat_storage_controller_devices()
        grouped_devices = inst._join_block_devices(scsi_devices)
        for group in grouped_devices:
            device = MockInfiniBoxMultipathStorageController(group)
            multipath_devices.append(device)
        return multipath_devices

    @staticmethod
    def get_all_storage_controller_devices(inst, *args, **kwargs):
        return []

    @staticmethod
    def get_volume_type(inst, *args, **kwargs):
        from infi.vendata.vmware_powertools.utils import get_system_object_by_device_vendor
        system = get_system_object_by_device_vendor(inst)
        volume = system.volumes.get(id=inst.get_volume_id())
        return volume.get_type()

    def get_display_name_function(self):
        def get_display_name(inst, *args, **kwargs):
            from infi.vendata.smock.storagemodel.multipath import MockMultipathBlockDevice
            from infi.vendata.smock.storagemodel.multipath import MockMultipathStorageController
            serial_number = inst.get_scsi_serial_number()
            if serial_number in self._vmock._device_names:
                return self._vmock._device_names[serial_number]
            elif isinstance(inst, MockMultipathStorageController):
                return "Storage Controller Array {}".format(serial_number)
            elif isinstance(inst, MockMultipathBlockDevice):
                return "Fibre Channel Disk {}".format(serial_number)
            return serial_number
        return get_display_name

    def _smock(self):
        from infi.vendata.smock.storagemodel.multipath import MockMultipathFrameworkModel
        from infi.vendata.smock.storagemodel.multipath import MockMultipathBlockDevice, MockMultipathStorageController
        from infi.vendata.smock.storagemodel.scsi import MockSCSIModel
        from infi.storagemodel.vendor.infinidat.infinibox.mixin.volume import InfiniBoxVolumeMixin
        patchers = [
            patch.object(MockSCSIModel, "get_all_storage_controller_devices",
                         self.get_all_storage_controller_devices),
            patch.object(MockMultipathFrameworkModel, "get_all_multipath_storage_controller_devices",
                         self.get_all_multipath_storage_controller_devices),
            patch.object(MockMultipathBlockDevice, "get_canonical_name", self.get_canonical_name, create=True),
            patch.object(MockMultipathBlockDevice, "get_uuid", self.get_uuid, create=True),
            patch.object(MockMultipathStorageController, "get_uuid", self.get_uuid, create=True),
            patch.object(MockMultipathBlockDevice, "get_display_name", self.get_display_name_function()),
            patch.object(MockMultipathStorageController, "get_display_name", self.get_display_name_function()),
            patch.object(InfiniBoxVolumeMixin, "get_volume_type", self.get_volume_type),
        ]
        [patcher.start() for patcher in patchers]
        self.patchers += patchers

    def start(self):
        self._smock()

class PyvisdkPatches(PatchesCollection):  # pragma: no cover
    def start(self):
        self._vim()
        self._sms()
        self._esxcli()

    def _vim(self):
        patcher = patch("infi.pyvmomi_wrapper.Client")
        Vim = patcher.start()
        Vim.return_value = self._vmock.get_client()
        self.patchers.append(patcher)

    def _sms(self):
        patcher = patch("infi.pyvmomi_wrapper.sms.SmsClient")
        patcher.start()
        #Sms.return_value = self._vmock.get_client()
        self.patchers.append(patcher)

    def _esxcli(self):
        patcher = patch("infi.pyvmomi_wrapper.esxcli.EsxCLI")
        patcher.start()
        self.patchers.append(patcher)

class IntegrationTestsPatches(PatchesCollection):  # pragma: no cover
    def start(self):
        self._purge_all_related_zones()
        self._allocate_infinidat_system()
        self._zone_host_with_system__full_mesh()

    def _purge_all_related_zones(self):
        from . import vMockFactory
        from infi.storagemodel.vmware.patches.storagemodel import StorageModelFactory
        from infi.vendata.integration_tests.zoning import ZoneFactory
        patcher = patch.object(ZoneFactory, "purge_all_related_zones")
        def side_effect(*args, **kwargs):
            vMockFactory.purge_all_related_zones_on_host(StorageModelFactory.get()._host)
        purge_all_related_zones = patcher.start()
        purge_all_related_zones.side_effect = side_effect
        self.patchers.append(patcher)

    def _zone_host_with_system__full_mesh(self):
        from . import vMockFactory
        from infi.storagemodel.vmware.patches.storagemodel import StorageModelFactory
        from infi.vendata.integration_tests.zoning import ZoneFactory
        patcher = patch.object(ZoneFactory, "zone_host_with_system__full_mesh")
        def side_effect(*args, **kwargs):
            allocated_system, = args
            infinibox = allocated_system.get_infinisdk()
            vMockFactory.zone_host_with_infinibox__full_mesh(StorageModelFactory.get()._host, infinibox)
        zone_host_with_system__full_mesh = patcher.start()
        zone_host_with_system__full_mesh.side_effect = side_effect
        self.patchers.append(patcher)

    def _allocate_infinidat_system(self):
        from . import vMockFactory
        from infi.vendata.integration_tests.system_allocation import SystemFactory
        patcher = patch.object(SystemFactory, "allocate_infinidat_system")
        allocate_infinidat_system = patcher.start()
        def side_effect(*args, **kwargs):
            allocated_system = MagicMock()
            infinibox, simulator = vMockFactory.create_infinibox(self._vmock)
            allocated_system.get_infinisdk.return_value = infinibox
            def release(*args, **kwargs):
                for host in self._vmock.get_client().get_host_systems():
                    vMockFactory.clear_infinibox_and_host_relationship(host, infinibox)
                simulator.deactivate()
            allocated_system.release.side_effect = release
            return allocated_system
        allocate_infinidat_system.side_effect = side_effect
        self.patchers.append(patcher)
