import cherrypy
import re

from logging import getLogger
from infi.pyutils.lazy import cached_method
from mako.lookup import TemplateLookup
from ... import TEMPLATE_PATHS
from ....config import Configuration

logger = getLogger(__name__)

VMWARE_CONTEXT = {
                  'moref': r'^[A-Za-z0-9:\.\-]+$',
                  'locale': r'^[A-Za-z_]+$',
                  'vimSessionKey': r'^[a-zA-Z0-9\-]+$',
                  }

class ActiveSessions(object):
    sessions = {}
    usernames = {}

    @classmethod
    def get_counter(cls):
        return Configuration.get_active_session_count()

    @classmethod
    def purge(cls):
        cls.sessions = {}

class Privileges(object):
    counter = 100
    morefs_by_session_id = {}

    @classmethod
    def purge(cls):
        cls.morefs_by_session_id = {}

def get_vcenter(request):  # pragma: no cover
    return request.app.root.get_vcenter()

def get_vcenter_client(request):
    return get_vcenter(request).get_extension_client()

def get_session_id(request):
    return request.params['vimSessionKey']

def assert_syntax_of_arguments(request):
    kwargs = request.params
    for key, pattern in VMWARE_CONTEXT.items():
        value = kwargs.get(key)
        if value is None:
            raise cherrypy.HTTPError(400, 'Invalid request! Please restart your VMWare vSphere Client.')
        regex = re.compile(pattern)
        result = regex.match(value)
        if result is None:
            raise cherrypy.HTTPError(400, 'Invalid request! Please restart your VMWare vSphere Client.')

def get_extension_session_id(request):
    client = get_vcenter_client(request)
    return client.session_manager.currentSession.key

def get_active_sessions_from_vcenter(request):
    vcenter = get_vcenter(request)
    sessions = vcenter.get_active_sessions()
    return sessions

def assert_active_session_id(request):
    session_id = get_session_id(request)
    if ActiveSessions.sessions.get(session_id, 0) > 0:
        ActiveSessions.sessions[session_id] -= 1
        if ActiveSessions.sessions[session_id] == -1:
            ActiveSessions.usernames.pop(session_id, None)
        return
    sessions = get_active_sessions_from_vcenter(request)
    if session_id in sessions or session_id == get_extension_session_id(request):
        ActiveSessions.sessions[session_id] = ActiveSessions.get_counter()
        ActiveSessions.usernames[session_id] = sessions.get(session_id, None)
        return
    raise cherrypy.HTTPError(400, "Session lost! Please restart your VMWare vSphere Client.")

def check_for_privileges_on_managed_object(client, managed_object, session_id, privileges):
    checker = client.service_content.authorizationManager.HasPrivilegeOnEntity
    return checker(managed_object, session_id, privileges)

def check_for_privileges_in_vcenter(request, session_id, moref):
    client = get_vcenter_client(request)
    mobj = client.get_managed_object_by_reference(moref)
    resource_method = request.handler.callable
    privileges = getattr(resource_method, 'PRIVILEGES', getattr(resource_method.im_class, 'PRIVILEGES', []))
    logger.info("Checking for privileges {} on moref {}".format(privileges, moref))
    actual = check_for_privileges_on_managed_object(client, mobj, session_id, privileges)
    logger.info("Result is: {}".format(actual))
    return all(actual)

def assert_privileges_on_managed_object(request):
    session_id = get_session_id(request)
    moref = request.params['moref']
    cached_privileges = Privileges.morefs_by_session_id
    value = 0
    if session_id in cached_privileges and moref in cached_privileges[session_id]:
        value = cached_privileges[session_id][moref]
    if value == 0:
        if not check_for_privileges_in_vcenter(request, session_id, moref):
            raise cherrypy.HTTPError(400, "Insufficient privileges")
        value = Privileges.counter
    else:
        value -= 1
    privileges = cached_privileges.get(session_id, {})
    privileges.update({moref: value})
    cached_privileges[session_id] = privileges

def assert_browser():
    from infi.vendata.powertools.gui.webapp.plugin.resource import is_supported_browser
    if not is_supported_browser(cherrypy.request.headers):
        logger.info("Unsupported browser. Redirecting to static page")
        raise cherrypy.HTTPRedirect("/static/vmware/html/unsupported.html")

_webapp_ready = False

def set_ready():
    global _webapp_ready
    logger.info("Server now ready, won't redirect to warmup page anymore")
    _webapp_ready = True

def assert_ready():
    global _webapp_ready
    from urllib import urlencode
    if _webapp_ready:
        return
    logger.info("Not ready yet. Redirecting to warmup page")
    querystring = urlencode(dict(url="{}?{}".format(cherrypy.request.path_info, cherrypy.request.query_string)))
    raise cherrypy.HTTPRedirect("/plugin/warmup?" + querystring)

def assert_request(request):
    assert_syntax_of_arguments(request)
    assert_active_session_id(request)
    assert_privileges_on_managed_object(request)

def get_request_id(request):
    return hash(request)

def vmware_vcenter_client_request_hook():
    request = cherrypy.request
    request_id = get_request_id(request)
    logger.info("Validating request {}".format(request_id))
    try:
        assert_request(request)
        logger.info("Request {} is valid".format(request_id))
    except:
        logger.exception("Request {} is not valid".format(request_id))
        raise

def redirect_vwc():
    # our VWC java plugin redirects to our server with the X-Forwarded-* headers. If we see this header,
    # we know that VWC made the request, so we need to translate the URIs in the body of the response for it
    if cherrypy.request.headers.get("X-Forwarded-For", None) is None:
        return
    from infi.vendata.vmware_powertools.vwc.register_vwc import sanitized_key
    vwc_prefix = "/vsphere-client/" + sanitized_key() + "/proxy"

    # see before_finalize documentation and cherrypy/lib/caching.py:tee_output
    def wrap(body):
        for chunk in body:
            chunk = chunk.replace("'/plugin", "'" + vwc_prefix + "/plugin")
            chunk = chunk.replace('"/plugin', '"' + vwc_prefix + "/plugin")
            chunk = chunk.replace("'/static", "'" + vwc_prefix + "/static")
            chunk = chunk.replace('"/static', '"' + vwc_prefix + "/static")
            yield chunk

    cherrypy.response.body = wrap(cherrypy.response.body)

def set_vmware_vcenter_client_request_hook():
    cherrypy.tools.vmware_vcenter_client_request = cherrypy.Tool('before_handler', vmware_vcenter_client_request_hook)
    cherrypy.tools.assert_browser = cherrypy.Tool('before_handler', assert_browser)
    cherrypy.tools.assert_ready = cherrypy.Tool('before_handler', assert_ready)
    cherrypy.tools.redirect_vwc = cherrypy.Tool('before_finalize', redirect_vwc)

class PluginResource(object):
    cached_method = staticmethod(cached_method)

    PRIVILEGES = ["Host.Config.Storage"]

    template_lookup = TemplateLookup(directories=TEMPLATE_PATHS, cache_enabled=False)

    session_data = {}

    def get_url_from_request(self):
        return "{}?{}".format(cherrypy.request.path_info, cherrypy.request.query_string)

    def clear_cache(self):
        from infi.pyutils.lazy import clear_cache
        clear_cache(self)

    def clear_cached_entry(self, method, *args, **kwargs):
        from infi.pyutils.lazy import clear_cached_entry
        clear_cached_entry(method, *args, **kwargs)

    def set_cached_entry(self, value, method, *args, **kwargs):
        from infi.pyutils.lazy import _get_instancemethod_cache_entry
        method_id = method.__method_id__
        key = _get_instancemethod_cache_entry(method_id, *args, **kwargs)
        if not hasattr(self, "_cache"):
            self._cache = dict()
        self._cache[key] = value

    def set_session_data(self, key, value):
        """ Store a data item associated with the current user session """
        key = self._session_key(key)
        self.session_data[key] = value

    def get_session_data(self, key):
        """ Retrieve a data item associated with the current user session, or None if not found """
        key = self._session_key(key)
        return self.session_data.get(key)

    def del_session_data(self, key):
        """ Forget a data item associated with the current user session """
        key = self._session_key(key)
        self.session_data.pop(key, None)

    def _session_key(self, key):
        return (get_session_id(cherrypy.request), key)

    def next_task_step(self, key, object_creator_function):
        session_data = self.get_session_data(key)
        if session_data:
            task_object = session_data
        else:
            task_object = object_creator_function()
            task_object.start()
            self.set_session_data(key, task_object)
        try:
            progress = task_object.next()
        except:
            self.del_session_data(key)
            raise
        if progress is StopIteration:
            self.del_session_data(key)
        return 'ok'


class ExtensionResource(PluginResource):
    EXTENSION_POINT = None
    TITLE = {"en": "INFINIDAT"}
