import cherrypy
import logging
from infi.vendata.vmware_powertools.webapp.plugin.resource import PluginResource
from infi.vendata.vmware_powertools.webapp.plugin.views.host.operations import ProvisionMixin, HostResizeVolumeTask, TaskMixin
from infi.vendata.vmware_powertools.webapp.plugin.views.host.operations import create_vmfs_datastore_by_serial, mount_vmfs_datastore_by_serial
from infi.vendata.vmware_powertools.webapp.plugin.views.host.operations import HostRenameVolume, convert_size
from infi.vendata.vmware_powertools.webapp.plugin.views.utils import fix_serial, get_system_serial
from infi.vendata.vmware_powertools.webapp import json_response
from infi.vendata.vmware_powertools.vcenter.shortcuts import moref_to_object, object_to_moref
from infi.vendata.vmware_powertools.patches import with_host
from infi.vendata.vmware_powertools.vcenter.shortcuts import create_vcenter_task
from infi.vendata.vmware_powertools.mapping import get_all_mappings_in_cluster, good_cluster_hosts
from infi.vendata.vmware_powertools.utils import spawn, joinall

logger = logging.getLogger(__name__)


def cluster_datastores_changed(esx_cluster, system_serial):
    # Called when a datastore was added to the cluster or removed from it.
    # Clears the cache of the cluster's hosts' datastores.
    host_datastores = cherrypy.tree.apps[''].root.plugin.views.host.datastores
    for esx_host in good_cluster_hosts(esx_cluster):
        host_moref = object_to_moref(esx_host)
        host_datastores.clear_cached_entry(host_datastores.get_row_data, host_moref, system_serial)


class ClusterMapVolume(object):
    def _mount_datastore(self):
        self._task.set_description("Mounting datastore")
        joinall([spawn(operation.mount_datastore) for operation in self._operations], raise_error=True)

    def _map_volume(self):
        from infi.vendata.vmware_powertools.utils import refresh_storage_model_on_cluster
        self._task.set_description("Mapping volume to cluster")
        self._operations[0].map_to_cluster()                                      # map to cluster only once
        refresh_storage_model_on_cluster(self._esx_cluster)                       # but rescan on all hosts

    def run(self, esx_cluster, system_serial, volume_id, do_mount=True):
        from infi.vendata.vmware_powertools.vcenter.map import MapVolume
        self._task = create_vcenter_task(esx_cluster, "Volume map")
        self._esx_cluster = esx_cluster
        self._operations = [MapVolume(esx_host, int(system_serial), int(volume_id))
                            for esx_host in good_cluster_hosts(esx_cluster, maintenance_hosts=False)]
        steps = [self._map_volume, self._mount_datastore] if do_mount else [self._map_volume]
        self._task.step(steps)


class ClusterUnmapVolume(object):
    def __init__(self, esx_cluster, proper_serial):
        from infi.vendata.vmware_powertools.vcenter.unmap import VolumeUnmapBase
        self._esx_cluster = esx_cluster
        self._proper_serial = proper_serial
        self._operations = [VolumeUnmapBase(esx_host, proper_serial, True) for esx_host in good_cluster_hosts(esx_cluster, maintenance_hosts=False)]
        self._task = create_vcenter_task(esx_cluster, "Volume unmap")

    def _assert_cluster_mapping(self):
        from infi.vendata.vmware_powertools.utils import get_infinidat_device_by_serial
        from infi.vendata.vmware_powertools.mapping import is_cluster_mapped
        from infi.vendata.vmware_powertools.vcenter.unmap import ValidationError
        device = None
        for host in good_cluster_hosts(self._esx_cluster):
            device, matching_host = get_infinidat_device_by_serial(host, self._proper_serial), host
            if device is not None:
                break
        if device is None:
            raise ValidationError("Could not find a storage device matching the SCSI serial for this volume: {}".format(self._proper_serial))
        with with_host(matching_host):
            if not is_cluster_mapped(device):
                raise ValidationError("The volume is mapped to a specific host and not to the cluster, so it cannot be unmapped from this view.")

    def _check_assertions(self):
        self._task.set_description("Checking if lun can be safely detached")
        self._assert_cluster_mapping()
        joinall([spawn(operation.assert_safe_to_remove) for operation in self._operations], raise_error=True)

    def _unmount_datastore(self):
        self._task.set_description("Unmounting and detaching")
        joinall([spawn(operation.unmount_datastore) for operation in self._operations], raise_error=True)
        joinall([spawn(operation.detach) for operation in self._operations], raise_error=True)

    def _unmap(self):
        from infi.vendata.vmware_powertools.utils import refresh_storage_model_on_cluster
        self._task.set_description("Unmapping volume")
        self._operations[0].unmap()
        refresh_storage_model_on_cluster(self._esx_cluster)

    def _delete(self):
        self._task.set_description("Deleting volume")
        self._operations[0].delete()

    def unmap(self):
        steps = [self._check_assertions, self._unmount_datastore, self._unmap]
        self._task.step(steps)

    def delete(self):
        steps = [self._check_assertions, self._unmount_datastore, self._unmap, self._delete]
        self._task.step(steps)


class ClusterProvisionVolumeTask(ProvisionMixin, TaskMixin, object):
    def __init__(self, esx_cluster, system_serial, pool_name, volume_name, volume_size, thin, do_filesystem, set_device_name, volume_count=1):
        super(ClusterProvisionVolumeTask, self).__init__()
        self._esx_cluster = esx_cluster
        self._esx_object = esx_cluster
        self._system_serial = system_serial
        self._pool_name = pool_name
        self._volume_name = volume_name
        self._volume_size = volume_size
        self._volume_count = volume_count
        self._thin = thin
        self._do_filesystem = do_filesystem
        self._set_device_name = set_device_name
        self._task_name = "Volume provisioning"
        self._hosts = good_cluster_hosts(self._esx_cluster, maintenance_hosts=False)
        self._cluster = None
        self._volumes = []

    def _preparation(self):
        from random import choice
        # pick primary provisioning host
        self._esx_host = choice(self._hosts)
        logger.info("Picked ESX host '{}' as main provisioing host for cluster".format(self._esx_host.name))

        super(ClusterProvisionVolumeTask, self)._preparation()

    def _volume_map(self):
        self._task.set_description("Mapping volume{} to cluster" + self._suffix)
        self._cluster = self._system.host_clusters.get(id=self._system_info.get_cluster_id())
        [self._cluster.map_volume(volume) for volume in self._volumes]

    def _mount_on_other_hosts(self):
        def op(host):
            self._vmfs_mount(host)
        other_hosts = [host for host in self._hosts if host._moId != self._esx_host._moId]
        joinall([spawn(op, host) for host in other_hosts], raise_error=True)
        self._task.set_description("Waiting for mounts to appear")
        self._rescan_on_all_hosts(set_description=False)

    def _rescan_on_all_hosts(self, set_description=True):
        if set_description:
            self._task.set_description("Waiting for volume{} to appear".format(self._suffix))
        def op(host):
            self._wait_for_volume(host)
        joinall([spawn(op, host) for host in self._hosts], raise_error=True)

    def _finish(self):
        joinall([spawn(self._refresh_storagemodel, host) for host in self._hosts], raise_error=True)
        cluster_datastores_changed(self._esx_cluster, self._system_serial)
        return super(ClusterProvisionVolumeTask, self)._finish()

    def _get_steps(self):
        steps = [
            self._preparation,
            self._volume_create,
            self._volume_map,
            self._rescan_on_all_hosts
        ]
        if self._do_filesystem:
            steps.append(self._vmfs_create)
            steps.append(self._mount_on_other_hosts)
        steps.append(self._finish)
        return steps


class Operations(PluginResource):
    @json_response
    def provision_params(self, moref, serial, **kwargs):
        from ...host.operations import get_default_provision_params
        esx_cluster = moref_to_object(moref)
        return get_default_provision_params(good_cluster_hosts(esx_cluster, maintenance_hosts=False)[0], serial)    # we use the first host, there should be no difference

    @json_response
    def get_provision_tasks(self, moref, serial, volume_name, pool_name, volume_size, thin, do_filesystem, set_device_name, count, **kwargs):
        suffix = "s" if int(count) > 1 else ""
        series = " series" if int(count) > 1 else ""
        ret = [
                'Connecting to InfiniBox',
                'Creating the new {} volume{} "{}"'.format(volume_size, series, volume_name),
                'Mapping volume{} to host'.format(suffix),
                'Waiting for volume discovery, this might take a while'
        ]
        if do_filesystem:
            ret.append('Creating filesystem')
            ret.append('Mounting volume on all remaining hosts in cluster')
        ret.append('Performing housekeeping')
        return ret

    @json_response
    def validate_provision_params(self, moref, serial, volume_name, pool_name, volume_size, thin, do_filesystem, set_device_name, count, **kwargs):
        from ...host.operations import validate_provision_params
        esx_cluster = moref_to_object(moref)
        proper_serial = int(serial)
        host = good_cluster_hosts(esx_cluster, maintenance_hosts=False)[0]
        validate_provision_params(host, proper_serial, pool_name, volume_name, volume_size, count)

    @json_response
    def map_existing(self, moref, serial, volume_id, **kwargs):
        esx_cluster = moref_to_object(moref)
        ClusterMapVolume().run(esx_cluster, serial, volume_id)
        cluster_datastores_changed(esx_cluster, int(serial))

    @json_response
    def query_mount_after_map(self, serial, volume_id, **kwargs):
        return dict(has_fs=False)

    @json_response
    def unmap(self, moref, serial, **kwargs):  # pragma: no cover
        esx_cluster = moref_to_object(moref)
        proper_serial = fix_serial(serial)
        ClusterUnmapVolume(esx_cluster, proper_serial).unmap()
        cluster_datastores_changed(esx_cluster, get_system_serial(serial))

    @json_response
    def delete_vol(self, moref, serial, **kwargs):  # pragma: no cover
        esx_cluster = moref_to_object(moref)
        proper_serial = fix_serial(serial)
        ClusterUnmapVolume(esx_cluster, proper_serial).delete()
        cluster_datastores_changed(esx_cluster, get_system_serial(serial))

    @json_response
    def rescan(self, moref, **kwargs):
        from infi.vendata.vmware_powertools.utils import refresh_storage_model_on_cluster
        refresh_storage_model_on_cluster(moref_to_object(moref), raise_error=True, do_rescan=True, do_refresh=True)

    @json_response
    def rename(self, moref, serial, name, rename_datastore, rename_device, **kwargs):
        esx_cluster = moref_to_object(moref)
        proper_serial = fix_serial(serial)
        host_mappings = get_all_mappings_in_cluster(esx_cluster, proper_serial)
        host = host_mappings[0]
        HostRenameVolume().rename(host, proper_serial, name, rename_datastore, rename_device)
        cluster_datastores_changed(esx_cluster, get_system_serial(serial))

    @json_response
    def create_filesystem(self, moref, serial, **kwargs):
        esx_cluster = moref_to_object(moref)
        proper_serial = fix_serial(serial)
        first = True
        for esx_host in good_cluster_hosts(esx_cluster, maintenance_hosts=False):
            func = create_vmfs_datastore_by_serial if first else mount_vmfs_datastore_by_serial
            func(esx_host, proper_serial)
            first = False

    @json_response
    def get_resize_tasks(self, moref, device_name, **kwargs):
        esx_cluster = moref_to_object(moref)
        host = esx_cluster.host[0]
        proper_serial = fix_serial(device_name)
        proper_size = 0
        return HostResizeVolumeTask(host, proper_serial, proper_size).get_task_list()

    @json_response
    def run_resize_task(self, moref, device_name, size, **kwargs):
        # Note: device_name contains the volume serial
        esx_cluster = moref_to_object(moref)
        task_key = (HostResizeVolumeTask, moref)
        proper_serial = fix_serial(device_name)
        proper_size = convert_size(size)
        # Find a host to perform the resize with
        host_mappings = get_all_mappings_in_cluster(esx_cluster, proper_serial)
        host = host_mappings[0]
        task_object_creator = lambda: HostResizeVolumeTask(host, proper_serial, proper_size)
        return self.next_task_step(task_key, task_object_creator)

    @json_response
    def run_provision_task(self, moref, serial, volume_name, pool_name, volume_size, thin, do_filesystem, set_device_name, count, **kwargs):
        esx_cluster = moref_to_object(moref)
        task_key = (ClusterProvisionVolumeTask, moref)
        proper_serial = int(serial)
        proper_size = convert_size(volume_size)
        volume_count = int(count)
        task_object_creator = lambda: ClusterProvisionVolumeTask(esx_cluster, proper_serial, pool_name, volume_name, proper_size, thin, do_filesystem, set_device_name, volume_count)
        return self.next_task_step(task_key, task_object_creator)
