from logging import getLogger
from . import actions, queries
from infi.pyutils.lazy import cached_method
from infi.vendata.vmware_powertools.mapping import good_cluster_hosts
logger = getLogger(__name__)


class VMwareClusterRegistration(object):
    def __init__(self, esx_cluster, system_serial):
        super(VMwareClusterRegistration, self).__init__()
        self.esx_cluster = esx_cluster
        self.system_serial = system_serial
        self.cluster_object = None

    def register_cluster(self):
        from infi.vendata.powertools.system_register.engine import output_stream_context
        from StringIO import StringIO
        with output_stream_context(StringIO()):
            self.register_all_hosts_in_system()
            self.register_all_hosts_in_cluster()

    def register_all_hosts_in_cluster(self):
        for esx_host in good_cluster_hosts(self.esx_cluster):
            self.register_esx_host_in_cluster(esx_host)
        actions.refresh_storage_model(self.esx_cluster)

    def register_all_hosts_in_system(self):
        from infi.vendata.vmware_powertools.webapp.plugin.views.host import registration as host_registration
        from infi.vendata.vmware_powertools.mapping import good_cluster_hosts
        from infi.vendata.vmware_powertools.utils import spawn, joinall
        def op(esx_host):
            if not host_registration.is_host_registration_supported(esx_host, self.system_serial):
                return
            if not host_registration.is_host_registered_in_system(esx_host, self.system_serial):
                host_registration.register_host(esx_host, self.system_serial)
        joinall([spawn(op, esx_host) for esx_host in good_cluster_hosts(self.esx_cluster)], raise_error=True)
        actions.refresh_storage_model(self.esx_cluster)

    @cached_method
    def get_esx_cluster_name(self):
        return queries.get_esx_cluster_name(self.esx_cluster)

    @cached_method
    def get_system_object_from_inquiry_mixin(self, inquiry_mixin):
        return queries.get_infinisdk_system_object_from_inquiry(inquiry_mixin)

    def is_cluster_object_exists(self, system_object, expected_cluster_name):
        if self.cluster_object is not None:
            return True
        if queries.is_cluster_object_exists(system_object, expected_cluster_name):
            self.cluster_object = queries.get_cluster_object_by_name(system_object, expected_cluster_name)
            return True
        return False

    @cached_method
    def are_there_foreign_hosts_in_cluster(self, cluster_object):
        return queries.are_there_foreign_hosts_in_cluster(cluster_object.get_system(), cluster_object, self.esx_cluster)

    @cached_method
    def get_system_object(self):
        inquiry_mixin = queries.get_inquiry_mixin_for_system_by_serial(self.system_serial)
        return self.get_system_object_from_inquiry_mixin(inquiry_mixin)

    def register_esx_host_in_cluster(self, esx_host):
        from infi.vendata.vmware_powertools.patches import with_host
        from infinisdk.core.exceptions import ObjectNotFound
        with with_host(esx_host):
            system_object = self.get_system_object()
            expected_cluster_name = self.get_esx_cluster_name()
            try:
                host_object = system_object.hosts.get(name=esx_host.summary.config.name)
            except ObjectNotFound:
                inquiry_mixin = queries.get_inquiry_mixin_for_system_by_serial(self.system_serial)
                host_object = queries.get_host_object_in_system(system_object, inquiry_mixin)
            logger.debug("Registering host {} in cluster {}".format(host_object.get_name(), expected_cluster_name))
        if queries.is_host_part_of_a_cluster(host_object):
            logger.debug("Host is a member of a cluster")
            cluster_object = host_object.get_cluster()
            if queries.is_cluster_name_matches(cluster_object, expected_cluster_name):
                return
            else:
                if not self.are_there_foreign_hosts_in_cluster(cluster_object):
                    actions.rename_cluster(cluster_object, expected_cluster_name)
                    return
                luns = dict()
                if queries.are_there_volumes_mapped_to_cluster(cluster_object):
                    luns.update({lun: lun_mapping.volume
                                 for lun, lun_mapping in cluster_object.get_luns()
                                 if lun_mapping.clustered})
                    next_name = queries.find_next_available_name(system_object, expected_cluster_name)
                    actions.rename_cluster(cluster_object, next_name)
                actions.remove_host_from_cluster(cluster_object, host_object)
                if queries.is_cluster_empty_of_hosts(cluster_object):
                    actions.delete_cluster(cluster_object)
                logger.debug("Cluster object renamed, creating cluster")
                self.cluster_object = actions.create_cluster(system_object, expected_cluster_name)
                logger.debug("Adding host to cluster")
                actions.add_host_to_cluster(self.cluster_object, host_object)
                if luns:
                    mapped_volumes = [lun_mapping.volume for lun_mapping in self.cluster_object.get_luns()]
                    for lun, volume in [item for item in luns.items() if item[1] not in mapped_volumes]:
                        self.cluster_object.map_volume(volume=volume, lun=lun)
        else:
            logger.debug("Host is not a member of a cluster")
            if not self.is_cluster_object_exists(system_object, expected_cluster_name):
                logger.debug("Cluster object does not exist, creating cluster")
                self.cluster_object = actions.create_cluster(system_object, expected_cluster_name)
                logger.debug("Adding host to cluster")
                actions.add_host_to_cluster(self.cluster_object, host_object)
            else:
                logger.debug("Cluster object exists")
                if self.are_there_foreign_hosts_in_cluster(self.cluster_object):
                    logger.debug("There are foreign hosts members in the cluster object")
                    next_name = queries.find_next_available_name(system_object, expected_cluster_name)
                    logger.debug("Renaming cluster {} to {}".format(expected_cluster_name, next_name))
                    actions.rename_cluster(self.cluster_object, next_name)
                    logger.debug("Creating new cluster {}".format(expected_cluster_name))
                    self.cluster_object = actions.create_cluster(system_object, expected_cluster_name)
                logger.debug("There are no foreign hosts members in the cluster object")
                logger.debug("Adding host to cluster")
                actions.add_host_to_cluster(self.cluster_object, host_object)
