import cherrypy
import logging
import cjson
from munch import Munch
from ...resource import ExtensionResource
from infi.vendata.vmware_powertools.webapp import json_response
from infi.vendata.vmware_powertools.snapshots.vendor_query import VendorQuery
from infi.vendata.vmware_powertools.snapshots.snapshot_operations import SnapshotOperations
from infi.vendata.vmware_powertools.snapshots.snapshotstore import SnapshotStore
from infi.vendata.vmware_powertools.vcenter import vCenter
from infi.vendata.vmware_powertools.vcenter.shortcuts import moref_to_object, object_to_moref

logger = logging.getLogger(__name__)

class Datastore(ExtensionResource):
    EXTENSION_POINT = "InventoryView.Datastore"

    @cherrypy.expose
    def index(self, moref, **kwargs):
        return self.get_response(moref)

    def get_response(self, moref):
        from time import time
        from infi.vendata.vmware_powertools.__version__ import __version__
        url = self.get_url_from_request()
        sources = self._get_sources(moref, url)
        mako_kwargs = dict(sources=cjson.encode(sources))
        mako_kwargs["timestamp"] = time()
        mako_kwargs["powertools_version"] = __version__
        datastore_object = moref_to_object(moref)
        mako_kwargs["datastorename"] = datastore_object.name
        vendor_data = VendorQuery.get_datastore_infinidat_vendor(datastore_object)
        mako_kwargs["is_infinidat_datastore"] = vendor_data is not None
        if vendor_data is not None:
            mako_kwargs["infinibox_name"] = vendor_data.get_system_name()
            mako_kwargs["infinibox_host_name"] = vendor_data.get_host_name()
            mako_kwargs["infinibox_volume_name"] = vendor_data.get_volume_name()
        return self.template_lookup.get_template('datastore.mako').render(**mako_kwargs)

    def _get_url_to_vm(self, url, strip_moref=True):
        import re
        if strip_moref:
            # get rid of the datastore moref, because we need the reference to be for the VM (and that's done in JS)
            url = re.sub("(\?|&)moref=([a-zA-Z0-9-:]+)(&|$)",
                   lambda matchobj: '?' if matchobj.group(1) == '?' else matchobj.group(3),
                   url)
        url = url.replace("/views", "/redirect")
        url = url.replace("/datastore/?", "/virtualmachine?back_element=Datastore&")
        return url

    def _get_sources(self, moref, url):
        sources = Munch()
        url = url.replace("/redirect", "/views")
        sources.refresh = url.replace("/datastore/?", "/datastore/refresh?")
        sources.vm_table = url.replace("/datastore/?", "/datastore/vm_table?")
        sources.operations = Munch()
        sources.operations.goto_vm = self._get_url_to_vm(url)
        sources.operations.goto_dead_vm = self._get_url_to_vm(url, False)
        sources.operations.create_snapshot = url.replace("/datastore/?", "/datastore/create_snapshot?")
        sources.auth = Munch()
        sources.auth.has_credentials = url.replace("/datastore/?", "/auth/has_credentials/?")
        sources.auth.set_credentials = url.replace("/datastore/?", "/auth/set_credentials/?")
        sources.auth.reset_credentials = url.replace("/datastore/?", "/auth/reset_credentials/?")
        return sources

    @json_response
    def refresh(self, moref, **kwargs):
        pass

    def _get_datastore_vms(self, moref):
        datastore_object = moref_to_object(moref)
        link_format = '<a href="#" class="goto_vm" id="{}" data-name="{}" data-serial="{}">{}</a>'
        dead_link_format = '<a href="#" class="goto_dead_vm" id="{}">{}</a> ' \
                           '<span class="vm-subtext">(Deleted)</span>'
        multiple_datastore_vm_link_format = '<a href="#" class="goto_vm" id="{}">{}</a> ' \
                           '<span class="vm-subtext">(Multiple datastores used)</span>'
        res = []
        appended_uuids = []
        infiniboxes = dict()
        for vm in datastore_object.vm:
            if vm.summary.config.uuid is None:
                continue
            if VendorQuery.one_infinidat_datastore_and_more(vm):
                res.append([multiple_datastore_vm_link_format.format(object_to_moref(vm), vm.name), ''])
            else:
                # one infinidat datastore only...
                infinibox_name, infinibox_serial = VendorQuery.get_system_name_and_serial_from_datastore(vm)
                infiniboxes[infinibox_serial] = infinibox_name
                res.append([link_format.format(object_to_moref(vm), infinibox_name, infinibox_serial, vm.name),
                            '<input type="checkbox" class="vm_checkbox">'])
            appended_uuids.append(vm.summary.config.uuid)

        snapshot_list = SnapshotStore.get_from_store(SnapshotStore.DATASTORE_SNAP_STORE, datastore_object.info.url)
        for snapshot_info in snapshot_list:
            if snapshot_info["vm_uuid"] not in appended_uuids:
                res.append([dead_link_format.format(snapshot_info["vm_uuid"],
                                                    snapshot_info["vm_name"]),
                            ""])
                appended_uuids.append(snapshot_info["vm_uuid"])

        # for the credentials dialog, we need to have div's with id="infinibox-{serial}" and data-name={name}
        # those divs exist in the datacenter/cluster/host views, but here we need to append them as hidden divs
        # to the last item
        div_format = '<div style="display:none" id="infinibox-{}" data-name="{}"></div>'
        if len(infiniboxes) > 0 and len(res) > 0:
            res[-1][0] = res[-1][0] + "".join([div_format.format(serial, name) for serial, name in infiniboxes.items()])

        return res

    @json_response
    def vm_table(self, moref, **kwargs):
        data = self._get_datastore_vms(moref)
        return dict(data=data)

    @json_response
    def create_snapshot(self, moref, vmrefs, snap_name, snap_desc, consistent, consolidate, **kwargs):
        client = vCenter.get_extension_client()
        datastore_object = moref_to_object(moref)
        vmrefs = vmrefs.rstrip(';').split(';')
        vm_objects = [client.get_managed_object_by_reference(vmref) for vmref in vmrefs]
        active_vm_objects = [item for item in vm_objects if item.summary.config.uuid is not None]
        SnapshotOperations.create_snapshot(datastore_object, active_vm_objects, snap_name, snap_desc, consistent, consolidate)

    @json_response
    def register_snapshot(self, moref, vmrefs, snapshot_id, **kwargs):
        from infi.vendata.vmware_powertools.snapshots.exceptions import SnapshotException
        client = vCenter.get_extension_client()
        datastore_object = moref_to_object(moref)
        vmrefs = vmrefs.rstrip(';').split(';')
        vm_objects = [client.get_managed_object_by_reference(vmref) for vmref in vmrefs]
        active_vm_objects = [item for item in vm_objects if item.summary.config.uuid is not None]
        if not active_vm_objects:
            raise SnapshotException("no active virtual machines in {}".format(vmrefs))
        SnapshotOperations.register_snapshot(datastore_object, active_vm_objects, snapshot_id)

