import cherrypy
import logging
import cjson
from munch import Munch
from ...resource import ExtensionResource
from infi.vendata.vmware_powertools.webapp import json_response
from infi.vendata.vmware_powertools.snapshots.snapshot_operations import SnapshotOperations
from infi.vendata.vmware_powertools.snapshots.snapshotstore import SnapshotStore
from infi.vendata.vmware_powertools.snapshots.vendor_query import VendorQuery
from infi.vendata.vmware_powertools.vcenter import vCenter
from infi.vendata.vmware_powertools.vcenter.shortcuts import object_to_moref

logger = logging.getLogger(__name__)

class Folder(ExtensionResource):
    EXTENSION_POINT = "InventoryView.VirtualMachineFolder"

    @cherrypy.expose
    def index(self, moref, **kwargs):
        return self.get_response(moref)

    def get_response(self, moref):
        from time import time
        from infi.vendata.vmware_powertools.__version__ import __version__
        url = self.get_url_from_request()
        sources = self._get_sources(moref, url)
        folder_object = self._get_folder_object(moref)
        mako_kwargs = dict(sources=cjson.encode(sources))
        mako_kwargs["timestamp"] = time()
        mako_kwargs["powertools_version"] = __version__
        mako_kwargs["foldername"] = folder_object.name
        return self.template_lookup.get_template('folder.mako').render(**mako_kwargs)

    def _get_url_to_vm(self, url, strip_moref=True):
        import re
        if strip_moref:
            # get rid of the folder moref, because we need the reference to be for the VM (and that's done in JS)
            url = re.sub("(\?|&)moref=([a-zA-Z0-9-:]+)(&|$)",
                lambda matchobj: '?' if matchobj.group(1) == '?' else matchobj.group(3),
                url)
        url = url.replace("/views", "/redirect")
        url = url.replace("/folder/?", "/virtualmachine?back_element=Folder&")
        return url

    def _get_sources(self, moref, url):
        sources = Munch()
        url = url.replace("/redirect", "/views")
        sources.refresh = url.replace("/folder/?", "/folder/refresh?")
        sources.vm_table = url.replace("/folder/?", "/folder/vm_table?")
        sources.operations = Munch()
        sources.operations.goto_vm = self._get_url_to_vm(url)
        sources.operations.goto_dead_vm = self._get_url_to_vm(url, False)
        sources.operations.create_snapshot = url.replace("/folder/?", "/folder/create_snapshot?")
        sources.auth = Munch()
        sources.auth.has_credentials = url.replace("/folder/?", "/auth/has_credentials/?")
        sources.auth.set_credentials = url.replace("/folder/?", "/auth/set_credentials/?")
        sources.auth.reset_credentials = url.replace("/folder/?", "/auth/reset_credentials/?")
        return sources

    def _get_folder_object(self, moref):
        client = vCenter.get_extension_client()
        return client.get_managed_object_by_reference(moref)

    @json_response
    def refresh(self, moref, **kwargs):
        pass

    def _get_folder_vms(self, moref):
        from pyVmomi import vim
        folder_object = self._get_folder_object(moref)
        link_format = '<a href="#" class="goto_vm" id="{}" data-name="{}" data-serial="{}">{}</a>'
        dead_link_format = '<a href="#" class="goto_dead_vm" id="{}">{}</a> ' \
                           '<span class="vm-subtext">(Deleted)</span>'
        noinfi_format = '{} <span class="vm-subtext">(Not on INFINIDAT Datastore)</span>'
        multiple_datastore_vm_link_format = '<a href="#" class="goto_vm" id="{}">{}</a> ' \
                           '<span class="vm-subtext">(Multiple datastores used)</span>'
        res = []
        appended_uuids = []
        infiniboxes = dict()
        for vm in folder_object.childEntity:
            if not isinstance(vm, vim.VirtualMachine):
                continue
            if not VendorQuery.has_infinidat_datastore(vm):
                res.append([noinfi_format.format(vm.name), ""])
            elif VendorQuery.one_infinidat_datastore_and_more(vm):
                res.append([multiple_datastore_vm_link_format.format(object_to_moref(vm), vm.name), ''])
            else:
                # exactly one infinidat datastore
                infinibox_name, infinibox_serial = VendorQuery.get_system_name_and_serial_from_datastore(vm)
                infiniboxes[infinibox_serial] = infinibox_name
                res.append([link_format.format(object_to_moref(vm), infinibox_name, infinibox_serial, vm.name),
                            '<input type="checkbox" class="vm_checkbox">'])
                appended_uuids.append(vm.summary.config.uuid)

        snapshot_list = SnapshotStore.get_from_store(SnapshotStore.FOLDER_SNAP_STORE, object_to_moref(folder_object))
        for snapshot_info in snapshot_list:
            if snapshot_info["vm_uuid"] not in appended_uuids:
                res.append([dead_link_format.format(snapshot_info["vm_uuid"],
                                                    snapshot_info["vm_name"]),
                            ""])
                appended_uuids.append(snapshot_info["vm_uuid"])

        # for the credentials dialog, we need to have div's with id="infinibox-{serial}" and data-name={name}
        # those divs exist in the datacenter/cluster/host views, but here we need to append them as hidden divs
        # to the last item
        div_format = '<div style="display:none" id="infinibox-{}" data-name="{}"></div>'
        if len(infiniboxes) > 0 and len(res) > 0:
            res[-1][0] = res[-1][0] + "".join([div_format.format(serial, name) for serial, name in infiniboxes.items()])

        return res

    @json_response
    def vm_table(self, moref, **kwargs):
        data = self._get_folder_vms(moref)
        return dict(data=data)

    @json_response
    def create_snapshot(self, moref, vmrefs, snap_name, snap_desc, consistent, consolidate, **kwargs):
        client = vCenter.get_extension_client()
        folder_object = self._get_folder_object(moref)
        vmrefs = vmrefs.rstrip(';').split(';')
        vm_objects = [client.get_managed_object_by_reference(vmref) for vmref in vmrefs]
        SnapshotOperations.create_snapshot(folder_object, vm_objects, snap_name, snap_desc, consistent, consolidate)
