import cherrypy
import logging
from ....resource import PluginResource
from ...utils import fix_serial, get_system_serial
from infi.vendata.vmware_powertools.webapp import json_response
from infi.vendata.vmware_powertools.vcenter.shortcuts import moref_to_object, object_to_moref, wait_for_task
from infi.vendata.powertools.utils import round_to_block_size
from infi.vendata.vmware_powertools.patches import with_host
from infi.vendata.vmware_powertools.vcenter.shortcuts import create_vcenter_task
from infi.vendata.vmware_powertools.utils import get_infinidat_device_by_serial, get_infinisdk_host_and_volume_by_device
from infi.vendata.vmware_powertools.utils import spawn, joinall
from infi.vendata.vmware_powertools.mapping import get_all_host_mappings
from capacity import byte

logger = logging.getLogger(__name__)

def update_device_name(esx_host, device):
    """ Change the device name to the volume name if they are different. Returns True if a change was made. """
    actual = device.get_display_name()
    expected = getattr(device.get_vendor(), 'get_volume_name', device.get_vendor().get_system_name)()
    if actual == expected:
        return False
    uuid = device.get_uuid()
    logger.info("Updating display name for {!r} with uuid {!r} to {!r}".format(device, uuid, expected))
    esx_host.configManager.storageSystem.UpdateScsiLunDisplayName(uuid, expected)
    #device._scsi_lun_data_object.displayName = expected
    return True

def get_datastore_by_serial_number(host, serial_number):
    from infi.vendata.vmware_powertools.vcenter.shortcuts import iter_infinidat_datastores
    with with_host(host):
        for datastore, infinidat_devices in iter_infinidat_datastores(host):
            for device in infinidat_devices:
                if device.get_scsi_serial_number() == serial_number:
                    return datastore
    return None

def create_vmfs_datastore(esx_host, block_device, volume):
    # for easy mock patching
    datastore_system = esx_host.configManager.datastoreSystem
    device_path = block_device.get_block_access_path()
    create_options = datastore_system.QueryVmfsDatastoreCreateOptions(device_path, 5)[0]
    create_options.spec.vmfs.volumeName = volume.get_name()
    return datastore_system.CreateVmfsDatastore(create_options.spec)

def create_vmfs_datastore_with_metadata(esx_host, block_device, volume):
    with with_host(esx_host):
        new_datastore = create_vmfs_datastore(esx_host, block_device, volume)
        volume.set_metadata_from_dict(dict(filesystem_type="VMFS", mount_point=new_datastore.host[0].mountInfo.path))
        return new_datastore

def create_vmfs_datastore_by_serial(esx_host, proper_serial):
    from infi.vendata.vmware_powertools.vcenter.unmap import VolumeUnmapBase
    VolumeUnmapBase(esx_host, proper_serial).assert_not_being_used_as_rdm()
    device = get_infinidat_device_by_serial(esx_host, proper_serial)
    with with_host(esx_host):
        _, volume = get_infinisdk_host_and_volume_by_device(device)
        new_datastore = create_vmfs_datastore_with_metadata(esx_host, device, volume)
        datastore_changed(proper_serial)
        return new_datastore

def mount_vmfs_datastore(esx_host, block_device):
    # A newly discovered vmfs volume will be mounted unless, it has been explicitly unmounted
    # But we leave this for easy mock patching
    pass

def mount_vmfs_datastore_by_serial(esx_host, proper_serial):
    device = get_infinidat_device_by_serial(esx_host, proper_serial)
    mount_vmfs_datastore(esx_host, device)

def get_default_provision_params(esx_host, serial):
    from infi.vendata.vmware_powertools.vcenter.map import HostSystemByController
    from infi.vendata.powertools.utils.queries import get_pools
    import uuid
    import capacity
    serial = int(serial)
    infinisdk_system = HostSystemByController(esx_host, serial).get_system()
    pools = [dict(name=pool.get_name(),
                  size=pool.get_virtual_capacity() // capacity.byte,
                  free=pool.get_free_virtual_capacity() // capacity.byte) for pool in get_pools(infinisdk_system)]
    return dict(box_serial=serial,
                filesystems=[dict(name="VMFS")],
                selected_filesystem="VMFS",
                suggested_volume_name="volume_{}".format(str(uuid.uuid1()).replace("-", "")),
                suggested_mountpoint="",
                pools=pools)

def validate_pool(esx_host, system_serial, pool_name):
    from infi.vendata.powertools.gui.webapp.plugin.views.host.operations import ProvisionValidationException
    from infi.vendata.vmware_powertools.vcenter.map import HostSystemByController
    system_pools = HostSystemByController(esx_host, system_serial).get_system().pools.get_administered_pools()
    if pool_name not in [pool.get_name() for pool in system_pools]:
        raise ProvisionValidationException("Unavailable pool")

def validate_provision_params(esx_host, system_serial, pool_name, volume_name, volume_size, count):
    from infi.vendata.powertools.gui.webapp.plugin.views.host.operations import ProvisionValidationException
    from infi.vendata.powertools.gui.webapp.plugin.views.host.operations import validate_size
    from infi.vendata.powertools.cli.validate import validate_volume_name
    from docopt import DocoptExit
    with with_host(esx_host):
        try:
            validate_volume_name(volume_name, system_serial)
        except DocoptExit as e:
            raise ProvisionValidationException(e.message.replace(e.usage, "")[:-1])
    validate_size(volume_size)
    validate_pool(esx_host, system_serial, pool_name)
    if not count.isdigit():
        raise ProvisionValidationException("Volume count is invalid")
    if int(count) > 20 or int(count) <= 0:
        raise ProvisionValidationException("Volume count must be between 1 and 20")

def host_datastores_changed(moref, system_serial):
    # Called when a datastore was added to the host or removed from it.
    # Clear the cache of the host's datastores
    host_datastores = cherrypy.tree.apps[''].root.plugin.views.host.datastores
    host_datastores.clear_cached_entry(host_datastores.get_row_data, moref, system_serial)

def datastore_changed(serial):
    # Clear the caches of the hosts that are mapped to the datastore
    # Should be called under with_host
    host_datastores = cherrypy.tree.apps[''].root.plugin.views.host.datastores
    host_datastores.clear_cache_by_datastore(serial)

def convert_size(volume_size):
    from infi.vendata.powertools.cli.validate import convert_size as powertools_convert_size
    volume_size = volume_size.replace("GB", "GiB").replace("TB", "TiB")
    return powertools_convert_size(volume_size)


class HostRenameVolume(object):
    def _wait_for_volume_change_on_host(self, host, volume_serial, expected_name):
        from infi.storagemodel import get_storage_model
        from infi.vendata.vmware_powertools.utils import get_infinidat_device_by_serial__inner
        with with_host(host):
            def _name_changed_predicate():
                dev = get_infinidat_device_by_serial__inner(volume_serial)
                return True if dev is None else dev.get_vendor().get_volume_name() == expected_name
            get_storage_model().rescan_and_wait_for(_name_changed_predicate, do_rescan=False, do_refresh=True)

    def _wait_for_volume_name_change(self, volume_serial, expected_name):
        all_hosts = get_all_host_mappings(volume_serial)
        joinall([spawn(self._wait_for_volume_change_on_host, host, volume_serial, expected_name)
                 for host in all_hosts], raise_error=True)

    def _update_name_for_device_with_serial_number(self, host, serial_number):
        with with_host(host):
            from infi.storagemodel.vendor.infinidat.shortcuts import get_infinidat_block_devices_and_controllers
            from infi.storagemodel import get_storage_model
            for device in [device for device in get_infinidat_block_devices_and_controllers()
                           if device.get_scsi_serial_number() == serial_number]:
                updated = update_device_name(host, device)
                if updated:
                    get_storage_model().refresh()
                    datastore_changed(serial_number)

    def rename(self, host, proper_serial, name, rename_datastore, rename_device):
        from infi.vendata.vmware_powertools.utils import get_infinisdk_volume_by_device
        from infi.vendata.powertools.cli.validate import validate_name
        validate_name(name)
        datastore = get_datastore_by_serial_number(host, proper_serial)
        device = get_infinidat_device_by_serial(host, proper_serial)
        with with_host(host):
            volume = get_infinisdk_volume_by_device(device)
        old_name = volume.get_name()
        # Rename volume
        if name != old_name:
            volume.update_name(name)
            self._wait_for_volume_name_change(proper_serial, name)
        # Rename datastore
        if datastore and rename_datastore and datastore.name != name:
            task = datastore.Rename_Task(name)
            wait_for_task(task)
        # Rename device
        if rename_device:
            self._update_name_for_device_with_serial_number(host, proper_serial)
        # Refresh
        datastore_changed(proper_serial)


class HostMapVolume(object):
    def _mount_datastore(self):
        self._task.set_description("Mounting datastore")
        self._operation.mount_datastore()

    def _map_volume(self):
        from infi.vendata.vmware_powertools.utils import refresh_storage_model_on_host
        self._task.set_description("Mapping volume")
        self._operation.map_to_host()
        refresh_storage_model_on_host(self._esx_host)

    def map(self, esx_host, system_serial, volume_id):
        from infi.vendata.vmware_powertools.vcenter.map import MapVolume
        self._esx_host = esx_host
        self._task = create_vcenter_task(esx_host, "Volume map")
        self._operation = MapVolume(esx_host, int(system_serial), int(volume_id))
        steps = [self._map_volume, self._mount_datastore]
        self._task.step(steps)


class HostUnmapVolume(object):
    def __init__(self, esx_host, proper_serial):
        from infi.vendata.vmware_powertools.vcenter.unmap import VolumeUnmapBase
        self._esx_host = esx_host
        self._proper_serial = proper_serial
        self._task = create_vcenter_task(esx_host, "Volume unmap")
        self._operation = VolumeUnmapBase(esx_host, proper_serial)

    def _check_assertions(self):
        from infi.vendata.vmware_powertools.mapping import is_cluster_mapped
        from infi.vendata.vmware_powertools.vcenter.unmap import ValidationError
        self._task.set_description("Checking if lun can be safely detached")
        device = get_infinidat_device_by_serial(self._esx_host, self._proper_serial)
        with with_host(self._esx_host):
            if is_cluster_mapped(device):
                raise ValidationError("The volume is cluster-mapped, it cannot be unmapped from this host.")
        self._operation.assert_safe_to_remove()

    def _detach(self):
        self._task.set_description("Unmounting and detaching")
        self._operation.unmount_datastore()
        self._operation.detach()

    def _unmap(self):
        from infi.vendata.vmware_powertools.utils import refresh_storage_model_on_host
        self._task.set_description("Unmapping volume")
        self._operation.unmap()
        refresh_storage_model_on_host(self._esx_host)

    def _delete(self):
        self._task.set_description("Deleting volume")
        self._operation.delete()

    def unmap(self):
        steps = [self._check_assertions, self._detach, self._unmap]
        self._task.step(steps)

    def force_unmap(self):
        steps = [self._unmap]
        self._task.step(steps)

    def delete(self):
        steps = [self._check_assertions, self._detach, self._unmap, self._delete]
        self._task.step(steps)


class TaskMixin(object):
    # For some tasks like unmap and rename we use the function 'step' from pyvmomi_wrapper's Task object, which
    # runs all the steps that are needed to finish the task one after the other. But for other tasks like
    # provisioning and rename we want to have progress indication on the webpage (with actions_modal), and for that
    # we use this class.
    # 'start' spawns '_run' which spawns a greenlet for each step. This acheives the same thing asg Task.step,
    # except we have a greenlet that we can monitor for each step. 'next' is called by the client, and it only
    # waits for the next greenlet to end, so that the client's browser can get progress indication.

    # subclasses need to implement _get_steps and define the members _esx_object and _task_name for the VMWare task
    def _run(self, steps):
        from infi.vendata.vmware_powertools.utils import spawn
        percentage = 0
        increment = 100 / len(steps)
        with self._task:
            for step in steps:
                greenlet = spawn(step)
                self._queue.put(greenlet)
                greenlet.get()      # join + raise exception if there was any
                percentage += increment
                self._task.update_progress(percentage)
            self._task.update_progress(100)

    def start(self):
        from gevent.queue import Queue
        from infi.vendata.vmware_powertools.utils import spawn
        self._task = create_vcenter_task(self._esx_object, self._task_name)
        self._queue = Queue()
        return spawn(self._run, self._get_steps())

    def next(self):
        greenlet = self._queue.get()
        return greenlet.get()      # join + raise exception if there was any

    def _finish(self):
        return StopIteration       # we don't raise to avoid logging; 'next_task_step' checks return value for this


class ProvisionMixin(object):
    def _preparation(self):
        from infi.vendata.powertools.utils.queries import get_system_controller_by_serial
        from infi.vendata.vmware_powertools.utils import get_system_object_by_device
        self._suffix = "s" if self._volume_count > 1 else ""
        self._task.set_description("Connecting to InfiniBox")
        with with_host(self._esx_host):
            system_controller = get_system_controller_by_serial(self._system_serial)
            self._system_info = system_controller.get_vendor()
            self._system = get_system_object_by_device(system_controller)

    def _volume_create(self):
        self._task.set_description("Creating volume" + self._suffix)
        self._pool = self._system.pools.get(name=self._pool_name)
        size = round_to_block_size(self._volume_size // byte) * byte
        provisioning = "thin" if self._thin else "thick"
        for i in range(self._volume_count):
            name = "{}{:03d}".format(self._volume_name, i+1) if self._volume_count > 1 else self._volume_name
            volume = self._system.volumes.create(pool=self._pool, name=name, size=size, provisioning=provisioning)
            self._volumes.append(volume)

    def _volume_map(self):
        self._task.set_description("Mapping volume{} to host".format(self._suffix))
        self._host = self._system.hosts.get(id=self._system_info.get_host_id())
        [self._host.map_volume(volume) for volume in self._volumes]

    def _get_block_device_by_volume_id(self, volume_id):
        from infi.storagemodel.vendor.infinidat.shortcuts import get_infinidat_block_devices
        return [device for device in get_infinidat_block_devices()
                if device.get_vendor().get_system_serial() == self._system_serial and
                   device.get_vendor().get_volume_id() == volume_id][0]

    def _wait_for_volume(self, esx_host=None):
        from infi.storagemodel.predicates import PredicateList
        from infi.storagemodel.vendor.infinidat.predicates import InfinidatVolumeExists
        from infi.storagemodel import get_storage_model
        if esx_host is None:
            self._task.set_description("Waiting for volume{} to appear".format(self._suffix))
        predicates = [InfinidatVolumeExists(self._system_serial, volume.get_id()) for volume in self._volumes]
        with with_host(esx_host or self._esx_host):
            get_storage_model().rescan_and_wait_for(PredicateList(predicates))
            self._block_devices = [self._get_block_device_by_volume_id(volume.get_id()) for volume in self._volumes]
            if self._set_device_name:
                joinall([spawn(update_device_name, esx_host or self._esx_host, block_device)
                         for block_device in self._block_devices], raise_error=True)

    def _vmfs_create(self):
        self._task.set_description("Creating VMFS")
        with with_host(self._esx_host):
            joinall([spawn(create_vmfs_datastore_with_metadata, self._esx_host, block_device, volume)
                     for block_device, volume in zip(self._block_devices, self._volumes)], raise_error=True)

    def _vmfs_mount(self, esx_host=None):
        joinall([spawn(mount_vmfs_datastore, esx_host or self._esx_host, block_device)
                 for block_device in self._block_devices], raise_error=True)

    def _refresh_storagemodel(self, esx_host=None):
        with with_host(esx_host or self._esx_host):
            from infi.storagemodel import get_storage_model
            get_storage_model().refresh()


class HostProvisionVolumeTask(ProvisionMixin, TaskMixin, object):
    def __init__(self, esx_host, system_serial, pool_name, volume_name, volume_size, thin, do_filesystem, set_device_name, volume_count=1):
        super(HostProvisionVolumeTask, self).__init__()
        self._esx_host = esx_host
        self._esx_object = esx_host
        self._system_serial = system_serial
        self._pool_name = pool_name
        self._volume_name = volume_name
        self._volume_size = volume_size
        self._volume_count = volume_count
        self._thin = thin
        self._do_filesystem = do_filesystem
        self._set_device_name = set_device_name
        self._task_name = "Volume provisioning"
        self._host = None
        self._volumes = []

    def _finish(self):
        self._refresh_storagemodel()
        moref = object_to_moref(self._esx_host)
        host_datastores_changed(moref, self._system_serial)
        return super(HostProvisionVolumeTask, self)._finish()

    def _get_steps(self):
        steps = [
            self._preparation,
            self._volume_create,
            self._volume_map,
            self._wait_for_volume
        ]
        if self._do_filesystem:
            steps.append(self._vmfs_create)
        steps.append(self._finish)
        return steps


class HostResizeVolumeTask(TaskMixin, object):
    def __init__(self, esx_host, proper_serial, proper_size):
        super(HostResizeVolumeTask, self).__init__()
        self._esx_host = esx_host
        self._esx_object = esx_host
        self._proper_serial = proper_serial
        self._proper_size = proper_size
        self._original_size = None
        self._volume = None
        self._task_name = "Volume resize"

    def get_task_list(self):
        res = ['Resizing the volume',
               'Rescanning']
        datastore = get_datastore_by_serial_number(self._esx_host, self._proper_serial)
        if datastore is not None:
            res.extend(['Resizing the filesystem'])
        res.extend([
                'Rescanning all related hosts',
                'Performing housekeeping'])
        return res

    def _resize_datastore(self):
        from infi.vendata.powertools.gui.webapp.plugin.views.host.operations import ProvisionValidationException
        self._task.set_description("Resizing VMFS datastore")
        manager = self._esx_host.configManager.datastoreSystem
        datastore = get_datastore_by_serial_number(self._esx_host, self._proper_serial)
        available_resize_options = manager.QueryVmfsDatastoreExpandOptions(datastore)
        # we can either resize by adding an extent, or by expanding an extend; we want to expand
        if not available_resize_options:
            raise ProvisionValidationException("No VMFS expand options")
        expand_spec = available_resize_options[0].spec
        manager.ExpandVmfsDatastore(datastore, expand_spec)

    def _size_changed_predicate(self):
        with with_host(self._esx_host):
            device = get_infinidat_device_by_serial(self._esx_host, self._proper_serial)
            _, volume = get_infinisdk_host_and_volume_by_device(device)
            return device.get_size_in_bytes() == (self._proper_size // byte)

    def _rescan(self):
        self._task.set_description("Rescanning")
        with with_host(self._esx_host):
            from infi.storagemodel import get_storage_model
            get_storage_model().rescan_and_wait_for(self._size_changed_predicate, do_rescan=False, do_refresh=True)
            datastore_changed(self._proper_serial)

    def _rescan_other_hosts(self):
        self._task.set_description("Rescanning all related hosts")
        # We need to rescan on all hosts that are mapped to the volume, not just this host.
        def rescan(esx_host):
            with with_host(esx_host):
                from infi.storagemodel import get_storage_model
                get_storage_model().rescan_and_wait_for(do_rescan=False, do_refresh=True)
                datastore_changed(self._proper_serial)
        greenlets = [spawn(rescan, esx_host) for esx_host in get_all_host_mappings(self._proper_serial)
                     if esx_host._moId != self._esx_host._moId]
        # We don't really wait for all the hosts and we don't add a predicate, because some hosts may have problems
        # and may take a lot of time to rescan. So we do best-effort...
        # We do wait a short time so that this step will appear in the task description - this way the rescans have
        # a context from the user's perspective.
        joinall(greenlets, timeout=3)

    def _resize_volume(self):
        from infi.vendata.vmware_powertools.vcenter.unmap import VolumeUnmapBase
        self._task.set_description("Resizing volume")
        VolumeUnmapBase(self._esx_host, self._proper_serial).assert_not_being_used_as_rdm()
        with with_host(self._esx_host):
            device = get_infinidat_device_by_serial(self._esx_host, self._proper_serial)
            _, volume = get_infinisdk_host_and_volume_by_device(device)
        self._original_size = volume.get_size()
        volume.update_size(self._proper_size)
        self._volume = volume

    def _get_steps(self):
        res = [
            self._resize_volume,
            self._rescan]
        datastore = get_datastore_by_serial_number(self._esx_host, self._proper_serial)
        if datastore is not None:
            res.extend([self._resize_datastore])
        res.extend([
            self._rescan_other_hosts,
            self._finish])
        return res


class Operations(PluginResource):
    @json_response
    def provision_params(self, moref, serial, **kwargs):
        esx_host = moref_to_object(moref)
        return get_default_provision_params(esx_host, serial)

    @json_response
    def get_provision_tasks(self, moref, serial, volume_name, pool_name, volume_size, thin, do_filesystem, set_device_name, count, **kwargs):
        suffix = "s" if int(count) > 1 else ""
        series = " series" if int(count) > 1 else ""
        ret = [
                'Connecting to InfiniBox',
                'Creating the new {} volume{} "{}"'.format(volume_size, series, volume_name),
                'Mapping volume{} to host'.format(suffix),
                'Waiting for volume discovery, this might take a while'
        ]
        if do_filesystem:
            ret.append('Creating filesystem')
        ret.append('Performing housekeeping')
        return ret

    @json_response
    def validate_provision_params(self, moref, serial, volume_name, pool_name, volume_size, thin, do_filesystem, set_device_name, count, **kwargs):
        host = moref_to_object(moref)
        proper_serial = int(serial)
        validate_provision_params(host, proper_serial, pool_name, volume_name, volume_size, count)

    @json_response
    def map_existing(self, moref, serial, volume_id, **kwargs):
        esx_host = moref_to_object(moref)
        HostMapVolume().map(esx_host, serial, volume_id)
        host_datastores_changed(moref, int(serial))

    @json_response
    def query_mount_after_map(self, serial, volume_id, **kwargs):
        return dict(has_fs=False)

    @json_response
    def unmap(self, moref, serial, **kwargs):
        from infi.storagemodel import get_storage_model
        host = moref_to_object(moref)
        proper_serial = fix_serial(serial)
        HostUnmapVolume(host, proper_serial).unmap()
        with with_host(host):
            get_storage_model().refresh()
        host_datastores_changed(moref, get_system_serial(serial))

    @json_response
    def delete_vol(self, moref, serial, **kwargs):
        from infi.storagemodel import get_storage_model
        host = moref_to_object(moref)
        proper_serial = fix_serial(serial)
        HostUnmapVolume(host, proper_serial).delete()
        with with_host(host):
            get_storage_model().refresh()
        host_datastores_changed(moref, get_system_serial(serial))

    @json_response
    def rescan(self, moref, **kwargs):
        from infi.vendata.vmware_powertools.utils import refresh_storage_model_on_host
        refresh_storage_model_on_host(moref_to_object(moref), raise_error=True, do_rescan=True, do_refresh=True)

    @json_response
    def rename(self, moref, serial, name, rename_datastore, rename_device, **kwargs):
        host = moref_to_object(moref)
        proper_serial = fix_serial(serial)
        HostRenameVolume().rename(host, proper_serial, name, rename_datastore, rename_device)

    @json_response
    def create_filesystem(self, moref, serial, **kwargs):
        host = moref_to_object(moref)
        proper_serial = fix_serial(serial)
        create_vmfs_datastore_by_serial(host, proper_serial)

    @json_response
    def get_resize_tasks(self, moref, device_name, **kwargs):
        host = moref_to_object(moref)
        proper_serial = fix_serial(device_name)
        proper_size = 0
        return HostResizeVolumeTask(host, proper_serial, proper_size).get_task_list()

    @json_response
    def run_resize_task(self, moref, device_name, size, **kwargs):
        # Note: device_name contains the volume serial
        host = moref_to_object(moref)
        task_key = (HostResizeVolumeTask, moref)
        proper_serial = fix_serial(device_name)
        proper_size = convert_size(size)
        task_object_creator = lambda: HostResizeVolumeTask(host, proper_serial, proper_size)
        return self.next_task_step(task_key, task_object_creator)

    @json_response
    def run_provision_task(self, moref, serial, volume_name, pool_name, volume_size, thin, do_filesystem, set_device_name, count, **kwargs):
        host = moref_to_object(moref)
        task_key = (HostProvisionVolumeTask, moref)
        proper_serial = int(serial)
        proper_size = convert_size(volume_size)
        volume_count = int(count)
        task_object_creator = lambda: HostProvisionVolumeTask(host, proper_serial, pool_name, volume_name, proper_size, thin, do_filesystem, set_device_name, volume_count)
        return self.next_task_step(task_key, task_object_creator)
