import cherrypy
import logging
import cjson
from munch import Munch
from ...resource import ExtensionResource
from infi.vendata.vmware_powertools.vcenter.shortcuts import moref_to_object
from snapstore import SnapStore
from operations import Operations
from infi.vendata.vmware_powertools.snapshots.vendor_query import VendorQuery
from infi.vendata.vmware_powertools.webapp import json_response

logger = logging.getLogger(__name__)

class VirtualMachine(ExtensionResource):
    EXTENSION_POINT = "InventoryView.VirtualMachine"

    @cherrypy.expose
    def index(self, **kwargs):
        return self.get_response(**kwargs)

    def _get_params(self, kwargs):
        if "vm_name" in kwargs and "vm_uuid" in kwargs:
            vm_name = kwargs["vm_name"]
            vm_uuid = kwargs["vm_uuid"]
            return None, vm_name, vm_uuid
        else:
            vm_object = moref_to_object(kwargs["moref"])
            vm_name = vm_object.name
            vm_uuid = vm_object.summary.config.uuid
            return vm_object, vm_name, vm_uuid

    def get_response(self, **kwargs):
        from time import time
        from urllib import unquote
        from infi.vendata.vmware_powertools.__version__ import __version__
        url = self.get_url_from_request()
        vm_object, vm_name, vm_uuid = self._get_params(kwargs)
        sources = self._get_sources(vm_uuid, url)
        mako_kwargs = dict(sources=cjson.encode(sources))
        mako_kwargs["back_element"] = None if "back_element" not in kwargs else kwargs["back_element"]
        if "back_url" in kwargs:
            back_url = unquote(kwargs["back_url"]).replace("/views", "/redirect")
        else:
            back_url = None
        mako_kwargs["back_url"] = back_url
        mako_kwargs["timestamp"] = time()
        mako_kwargs["powertools_version"] = __version__
        mako_kwargs["vmname"] = vm_name
        mako_kwargs["object_exists"] = vm_object is not None
        if vm_object is not None:
            datastore, vendor_data = VendorQuery.get_vm_infinidat_vendor(vm_object)
            mako_kwargs["is_infinidat_datastore"] = VendorQuery.has_infinidat_datastore(vm_object)
            mako_kwargs["too_many_datastores"] = VendorQuery.is_too_many_datastores(vm_object)
            if vendor_data is not None:
                mako_kwargs["datastore_name"] = self._get_datastore_part(datastore)
                mako_kwargs["infinibox_name"] = vendor_data.get_system_name()
                mako_kwargs["infinibox_serial"] = vendor_data.get_system_serial()
                mako_kwargs["infinibox_host_name"] = vendor_data.get_host_name()
                mako_kwargs["infinibox_volume_name"] = vendor_data.get_volume_name()
        else:
                # dead VM. Serial and name are unused but must be defined for the mako
                mako_kwargs["infinibox_serial"] = ""
                mako_kwargs["infinibox_name"] = ""
        return self.template_lookup.get_template('vm.mako').render(**mako_kwargs)

    def _get_datastore_part(self, datastore):
        return datastore.summary.name

    def _get_sources(self, vm_uuid, url):
        sources = Munch()
        url = url.replace("/redirect", "/views")
        sources.refresh = url.replace("/virtualmachine/?", "/virtualmachine/refresh?")
        sources.snapshot_table = url.replace("/virtualmachine/?",
                                             "/virtualmachine/snapstore/virtualmachine?uuid={}&".format(vm_uuid))
        sources.operations = Munch()
        for op in ["create_snapshot",
                   "get_revert_snapshot_parameters",
                   "revert_snapshot",
                   "expose_snapshot",
                   "get_tree_data",
                   "remove_snapshot",
                   "edit_snapshot",
                   "confirm_remove_snapshot",
                   "volume_info"]:
            sources.operations[op] = url.replace("/virtualmachine/?",
                                                 "/virtualmachine/operations/{}?uuid={}&".format(op, vm_uuid))
        sources.auth = Munch()
        sources.auth.has_credentials = url.replace("/virtualmachine/?", "/auth/has_credentials/?")
        sources.auth.set_credentials = url.replace("/virtualmachine/?", "/auth/set_credentials/?")
        sources.auth.reset_credentials = url.replace("/virtualmachine/?", "/auth/reset_credentials/?")
        return sources

    @json_response
    def refresh(self, moref, **kwargs):
        from pyVmomi import vim
        from infi.vendata.vmware_powertools.patches import with_host
        self.clear_cache()
        vm_object = moref_to_object(moref)
        if not isinstance(vm_object, vim.VirtualMachine):
            # if the VM is dead, moref points to the referring folder/datastore so there is no host..
            return
        host = vm_object.summary.runtime.host
        with with_host(host):
            from infi.storagemodel import get_storage_model
            get_storage_model().refresh()

    operations = Operations()
    snapstore = SnapStore()
