import cherrypy
import logging
from infi.vendata.vmware_powertools.vcenter import vCenter
from infi.vendata.vmware_powertools.snapshots.snapshot_operations import SnapshotOperations
from infi.vendata.vmware_powertools.snapshots.snapshotstore import SnapshotStore
from infi.vendata.vmware_powertools.webapp import json_response
from infi.vendata.vmware_powertools.vcenter.shortcuts import moref_to_object, object_to_moref

logger = logging.getLogger(__name__)

class Operations(object):
    def _get_group_list(self, uuid, snapshot_id):
        snapshot_info = SnapshotStore.get_snapshot_info(SnapshotStore.VM_SNAP_STORE, uuid, snapshot_id)
        snap_group_uuid = snapshot_info["snap_group_uuid"]
        snap_group_list = SnapshotStore.get_group_list(snap_group_uuid)
        return snap_group_list

    def _snapshots_exist(self, uuid):
        snapshot_list = SnapshotStore.get_from_store(SnapshotStore.VM_SNAP_STORE, uuid)
        return len(snapshot_list) > 0

    def _is_vm_deleted(self, moref):
        from pyVmomi import vim
        vm_object = moref_to_object(moref)
        # vm_object should be a VirtualMachine, but if the VM is deleted, moref will point to a datastore or folder
        return not isinstance(vm_object, vim.VirtualMachine)

    @json_response
    def create_snapshot(self, moref, snap_name, snap_desc, consistent, consolidate, **kwargs):
        vm_object = moref_to_object(moref)
        SnapshotOperations.create_snapshot(vm_object, [vm_object], snap_name, snap_desc, consistent, consolidate)

    @json_response
    def get_revert_snapshot_parameters(self, moref, uuid, snapshot_id, expose_only=False, **kwargs):
        from infi.vendata.vmware_powertools.snapshots.revert import get_default_revert_params
        snapshot_info = SnapshotStore.get_snapshot_info(SnapshotStore.VM_SNAP_STORE, uuid, snapshot_id)
        client = vCenter.get_extension_client()
        vm_object = None
        if not self._is_vm_deleted(moref):
            vm_object = client.get_managed_object_by_reference(moref)
        name, host, datastore, folder = get_default_revert_params(client, vm_object, snapshot_info)
        datacenter = client.get_datacenters()[0].parent if host is None else host.parent.parent.parent
        obj_to_data = lambda obj: None if obj is None else dict(moref=object_to_moref(obj), name=obj.name)
        return dict(name=name,
                    host=obj_to_data(host),
                    datastore=obj_to_data(datastore),
                    folder=obj_to_data(folder),
                    datacenter=obj_to_data(datacenter),
                    snapshot_id=snapshot_id,
                    expose_only=expose_only)

    @json_response
    def revert_snapshot(self, moref, uuid, snapshot_id, vm_name, host, datastore, folder, expose_only=False, **kwargs):
        vm_object = moref_to_object(moref)
        is_vm_deleted = self._is_vm_deleted(moref)
        host = moref_to_object(host)
        datastore = moref_to_object(datastore)
        folder = moref_to_object(folder)
        vm_name, new_datastore_name, cloned_vol_name = SnapshotOperations.revert_snapshot(vm_object, uuid, snapshot_id, not is_vm_deleted, vm_name, host, datastore, folder, expose_only)
        return dict(vm_name=vm_name, new_datastore_name=new_datastore_name, cloned_vol_name=cloned_vol_name, go_back=is_vm_deleted)   # if we revert a deleted VM, go back to parent

    @cherrypy.expose
    def get_tree_data(self, tree_moref, motype_name, toplevel=False, **kwargs):
        from .treeview import treeview_data
        import json
        client = vCenter.get_extension_client()
        return json.dumps(treeview_data(client, tree_moref, motype_name, toplevel))

    @json_response
    def confirm_remove_snapshot(self, moref, uuid, snapshot_id, **kwargs):
        snap_group_list = self._get_group_list(uuid, snapshot_id)
        snapshot_info = SnapshotStore.get_snapshot_info(SnapshotStore.VM_SNAP_STORE, uuid, snapshot_id)
        vm_snapshot_list = SnapshotStore.get_from_store(SnapshotStore.VM_SNAP_STORE, uuid)
        base_msg = "Are you sure you want to delete snapshot '{}'?".format(snapshot_info["snapshot_name"])
        if self._is_vm_deleted(moref) and len(vm_snapshot_list) == 1:
            return ("This is the last snapshot of a deleted VM. " + base_msg,
                    "Deleting this snapshot will cause the Virtual Machine to be unrecoverable.")
        for store, key, snapshot_id in snap_group_list:
            cur_snapshot_info = SnapshotStore.get_snapshot_info(store, key, snapshot_id)
            if cur_snapshot_info["vm_uuid"] != snapshot_info["vm_uuid"]:
                return ("This is a group snapshot. " + base_msg,
                        "Deleting it will delete the snapshot of other Virtual Machines as well.")
        return (base_msg, None)

    @json_response
    def remove_snapshot(self, moref, uuid, snapshot_id, **kwargs):
        snapshot_info = SnapshotStore.get_snapshot_info(SnapshotStore.VM_SNAP_STORE, uuid, snapshot_id)
        SnapshotOperations.remove_snapshot(snapshot_info)

        if not self._snapshots_exist(uuid) and self._is_vm_deleted(moref):
            # we just deleted the last snapshot of a deleted VM. signal UI to go back to parent object
            go_back = True
        else:
            go_back = False
        return dict(go_back=go_back)

    @json_response
    def edit_snapshot(self, moref, uuid, snapshot_id, **kwargs):
        snapshot_info = SnapshotStore.get_snapshot_info(SnapshotStore.VM_SNAP_STORE, uuid, snapshot_id)
        if 'name' in kwargs:
            name = kwargs['name'].strip()
            if not name:
                raise Exception('Snapshot name cannot be empty')
            snapshot_info['snapshot_name'] = name
        if 'desc' in kwargs:
            snapshot_info['snapshot_description'] = kwargs['desc']
        SnapshotOperations.update_snapshot(snapshot_info)

    @json_response
    def volume_info(self, moref, uuid, snapshot_id, **kwargs):
        from infi.vendata.vmware_powertools.snapshots.snapshot_operations import get_infinisdk_snapshot_volume
        snapshot_info = SnapshotStore.get_snapshot_info(SnapshotStore.VM_SNAP_STORE, uuid, snapshot_id)
        infinisdk_volume = get_infinisdk_snapshot_volume(snapshot_info)
        return dict(vm_list=infinisdk_volume.get_metadata_value('vm_list').split(","),
                    datastore_name=infinisdk_volume.get_metadata_value('original_datastore'),
                    snapshot_volume_name=infinisdk_volume.get_name(),
                    original_volume_name=infinisdk_volume.get_parent().get_name(),
                    snapshot_volume_size=str(infinisdk_volume.get_used_size()).replace('*', ' '))
